import re
import subprocess
import tempfile
import unittest
from fnmatch import fnmatch
import multiprocessing
from concurrent.futures import ProcessPoolExecutor
from time import sleep, time
import itertools
from pathlib import Path

import numpy as np

import sys
import errno
import os

from vmaf import run_process
from vmaf.tools.scanf import sscanf, IncompleteCaptureError, FormatError

__copyright__ = "Copyright 2016-2020, Netflix, Inc."
__license__ = "BSD+Patent"


try:
    multiprocessing.set_start_method('fork')
except ValueError:  # noqa, If platform does not support, just ignore
    pass
except RuntimeError:  # noqa, If context has already being set, just ignore
    pass


def get_stdout_logger():
    import logging
    logger = logging.getLogger()
    handler = logging.StreamHandler(stream=sys.stdout)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(handler)
    return logger


def close_logger(logger):
    for handler in logger.handlers:
        handler.close()
        logger.removeHandler(handler)


def get_file_name_without_extension(path):
    """

    >>> get_file_name_without_extension('yuv/src01_hrc01.yuv')
    'src01_hrc01'
    >>> get_file_name_without_extension('yuv/src01_hrc01')
    'src01_hrc01'
    >>> get_file_name_without_extension('abc/xyz/src01_hrc01.yuv')
    'src01_hrc01'
    >>> get_file_name_without_extension('abc/xyz/src01_hrc01.sdr.yuv')
    'src01_hrc01.sdr'
    >>> get_file_name_without_extension('abc/xyz/src01_hrc01.sdr.dvi.yuv')
    'src01_hrc01.sdr.dvi'

    """
    return Path(path).stem


def get_file_name_with_extension(path):
    """

    >>> get_file_name_with_extension('yuv/src01_hrc01.yuv')
    'src01_hrc01.yuv'
    >>> get_file_name_with_extension('src01_hrc01.yuv')
    'src01_hrc01.yuv'
    >>> get_file_name_with_extension('abc/xyz/src01_hrc01.yuv')
    'src01_hrc01.yuv'

    """
    return Path(path).name


def get_file_name_extension(path):
    """
    >>> get_file_name_extension("file:///mnt/zli/test.txt")
    'txt'
    >>> get_file_name_extension("test.txt")
    'txt'
    >>> get_file_name_extension("abc")
    ''
    >>> get_file_name_extension("test.265")
    '265'
    """
    return Path(path).suffix[1:]


def get_dir_without_last_slash(path: str) -> str:
    """

    >>> get_dir_without_last_slash('abc/src01_hrc01.yuv')
    'abc'
    >>> get_dir_without_last_slash('src01_hrc01.yuv')
    ''
    >>> get_dir_without_last_slash('abc/xyz/src01_hrc01.yuv')
    'abc/xyz'
    >>> get_dir_without_last_slash('abc/xyz/')
    'abc/xyz'

    """
    return os.path.dirname(path)


def make_parent_dirs_if_nonexist(path):
    dst_dir = get_dir_without_last_slash(path)
    os.makedirs(dst_dir, exist_ok=True)


def delete_dir_if_exists(dir_to_delete):
    if os.path.isdir(dir_to_delete):
        os.rmdir(dir_to_delete)


def get_normalized_string_from_dict(d):
    """ Normalized string representation with sorted keys.

    >>> get_normalized_string_from_dict({"max_buffer_sec": 5.0, "bitrate_kbps": 45, })
    'bitrate_kbps_45_max_buffer_sec_5.0'
    """
    return '_'.join(map(lambda k: '{k}_{v}'.format(k=k,v=d[k]), sorted(d.keys())))


def get_hashable_value_tuple_from_dict(d):
    """ Hashable tuple of values with sorted keys.

    >>> get_hashable_value_tuple_from_dict({"max_buffer_sec": 5.0, "bitrate_kbps": 45, })
    (45, 5.0)
    >>> get_hashable_value_tuple_from_dict({"max_buffer_sec": 5.0, "bitrate_kbps": 45, "resolutions": [(740, 480), (1920, 1080), ]})
    (45, 5.0, ((740, 480), (1920, 1080)))
    """
    return tuple(map(
        lambda k: tuple(d[k]) if isinstance(d[k], list) else d[k],
        sorted(d.keys())))


def get_unique_str_from_recursive_dict(d):
    """ String representation with sorted keys and values for recursive dict.

    >>> get_unique_str_from_recursive_dict({'a':1, 'b':2, 'c':{'x':'0', 'y':'1'}})
    '{"a": 1, "b": 2, "c": {"x": "0", "y": "1"}}'
    >>> get_unique_str_from_recursive_dict({'a':1, 'c':2, 'b':{'y':'1', 'x':'0', }})
    '{"a": 1, "b": {"x": "0", "y": "1"}, "c": 2}'
    """
    from collections import OrderedDict
    import json

    def to_ordered_dict_recursively(d):
        if isinstance(d, dict):
            return OrderedDict(map(
                lambda t: (to_ordered_dict_recursively(t[0]), to_ordered_dict_recursively(t[1])),
                sorted(d.items())
            ))
        else:
            return d
    return json.dumps(to_ordered_dict_recursively(d))


def indices(a, func):
    """
    Get indices of elements in an array which satisfies func
    >>> indices([1, 2, 3, 4], lambda x: x>2)
    [2, 3]
    >>> indices([1, 2, 3, 4], lambda x: x==2.5)
    []
    >>> indices([1, 2, 3, 4], lambda x: 1 < x <= 3)
    [1, 2]
    >>> indices([1, 2, 3, 4], lambda x: x in [2, 4])
    [1, 3]
    >>> indices([1,2,3,1,2,3,1,2,3], lambda x: x > 2)
    [2, 5, 8]
    """
    return [i for (i, val) in enumerate(a) if func(val)]


def import_python_file(filepath : str, override : dict = None):
    """
    Import a python file as a module, allowing overriding some of the variables.
    Assumption: in the original python file, variables to be overridden get assigned once only, in a single line.
    """
    if override is None:
        filename = get_file_name_without_extension(filepath)
        try:
            from importlib.machinery import SourceFileLoader
            ret = SourceFileLoader(filename, filepath).load_module()
        except ImportError:
            import imp
            ret = imp.load_source(filename, filepath)
        return ret
    else:
        override_ = override.copy()
        tmpfile = tempfile.NamedTemporaryFile(delete=False, suffix='.py')
        with open(filepath, 'r') as fin:
            with open(tmpfile.name, 'w') as fout:
                while True:
                    line = fin.readline()
                    if len(override_) > 0:
                        suffixes = []
                        for key in list(override_.keys()):
                            if key in line and '=' in line:
                                s = f"{key} = '{override_[key]}'" if isinstance(override_[key], str) else f"{key} = {override_[key]}"
                                suffixes.append(s)
                                del override_[key]
                        if len(suffixes) > 0:
                            line = '\n'.join([line.strip()] + suffixes) + '\n'
                    fout.write(line)
                    if not line:
                        break
                if len(override_) > 0:
                    for key in override_:
                        s = f"{key} = '{override_[key]}'" if isinstance(override_[key], str) else f"{key} = {override_[key]}"
                        s += '\n'
                        fout.write(s)
        #============= debug =================
        # with open(tmpfile.name, 'r') as fin:
        #     print(fin.read())
        #=====================================
        ret = import_python_file(tmpfile.name)
        os.remove(tmpfile.name)
        return ret


def make_absolute_path(path: str, current_dir: str) -> str:
    """
    >>> make_absolute_path('abc/cde.fg', '/xyz/')
    '/xyz/abc/cde.fg'
    >>> make_absolute_path('/abc/cde.fg', '/xyz/')
    '/abc/cde.fg'
    """
    assert current_dir.endswith('/'), f"expect current_dir ends with '/', but is: {current_dir}"
    if path[0] == '/':
        return path
    else:
        return current_dir + path


def empty_object():
    return type('', (), {})()


def get_cmd_option(argv, begin, end, option):
    """

    >>> get_cmd_option(['a', 'b', 'c', '--xyz', '123'], 3, 5, '--xyz')
    '123'
    >>> get_cmd_option(['a', 'b', 'c', '--xyz', '123'], 0, 5, '--xyz')
    '123'
    >>> get_cmd_option(['a', 'b', 'c', '--xyz', '123'], 4, 5, '--xyz')
    >>> get_cmd_option(['a', 'b', 'c', '--xyz', '123'], 5, 5, '--xyz')
    >>> get_cmd_option(['a', 'b', 'c', '--xyz', '123'], 6, 5, '--xyz')
    >>> get_cmd_option(['a', 'b', 'c', '--xyz', '123'], 0, 5, 'a')
    'b'
    >>> get_cmd_option(['a', 'b', 'c', '--xyz', '123'], 0, 5, 'b')
    'c'

    """
    itr = None
    for itr in range(begin, end):
        if argv[itr] == option:
            break
    if itr is not None and itr != end and (itr + 1) != end:
        return argv[itr + 1]
    return None


def cmd_option_exists(argv, begin, end, option):
    """

    >>> cmd_option_exists(['a', 'b', 'c', 'd'], 2, 4, 'c')
    True
    >>> cmd_option_exists(['a', 'b', 'c', 'd'], 3, 4, 'c')
    False
    >>> cmd_option_exists(['a', 'b', 'c', 'd'], 3, 4, 'd')
    True
    >>> cmd_option_exists(['a', 'b', 'c', 'd'], 2, 4, 'a')
    False
    >>> cmd_option_exists(['a', 'b', 'c', 'd'], 2, 4, 'b')
    False

    """
    found = False
    for itr in range(begin, end):
        if argv[itr] == option:
            found = True
            break
    return found


def index_and_value_of_min(l):
    """

    >>> index_and_value_of_min([2, 0, 3])
    (1, 0)

    """
    return min(enumerate(l), key=lambda x: x[1])

_pm_return_dict = None
_pm_func = None
_pm_list_args = None

def _parallel_map_rt(idx):
    _pm_return_dict[idx] = _pm_func(_pm_list_args[idx])

def parallel_map(func, list_args, processes=None):
    """
    Use multiprocessing.Pool to create a fast parallel map that doesn't pickle arguments
    Note: only works on Unix! fork() used to propagate unpickleable data
    """

    context = multiprocessing

    if getattr(multiprocessing, 'get_context', None) is None:
        assert os.name == 'posix', "parallel_map() requires fork() support, but not running on Unix"
    else:
        assert 'fork' in multiprocessing.get_all_start_methods(), "parallel_map() requires fork() support"
        context = multiprocessing.get_context('fork')

    # create shared dictionary
    return_dict = context.Manager().dict()

    def pool_init():
        global _pm_func, _pm_list_args, _pm_return_dict
        _pm_func = func
        _pm_list_args = list_args
        _pm_return_dict = return_dict

    with ProcessPoolExecutor(processes, mp_context=context, initializer=pool_init) as pool:
        # ProcessPoolExecutor prevents hanging on the slowest processes that get too much work - delegates one at a time
        for _ in pool.map(_parallel_map_rt, range(len(list_args))):
            pass
    
    return [return_dict[i] for i in range(len(list_args))]


def check_program_exist(program):
    """

    >>> check_program_exist("xxxafasd34df")
    False
    >>> check_program_exist("xxxafasd34df f899")
    False
    >>> check_program_exist("ls")
    True
    >>> check_program_exist("ls -all")
    True
    >>> check_program_exist("pwd")
    True

    """
    try:
        with open(os.devnull, "wb") as devnull_fd:
            subprocess.call(program.split(), stdout=devnull_fd)
        return True
    except OSError as e:
        if e.errno == errno.ENOENT:
            return False
        else:
            # Something else went wrong while trying to run `wget`
            raise


def check_scanf_match(string, template):
    """
    >>> check_scanf_match('frame00000000.icpf', 'frame%08d.icpf')
    True
    >>> check_scanf_match('frame00000003.icpf', 'frame%08d.icpf')
    True
    >>> check_scanf_match('frame0000001.icpf', 'frame%08d.icpf')
    True
    >>> check_scanf_match('frame00000001.icpff', 'frame%08d.icpf')
    True
    >>> check_scanf_match('gframe00000001.icpff', 'frame%08d.icpf')
    False
    >>> check_scanf_match('fyrame00000001.icpff', 'frame%08d.icpf')
    False
    >>> check_scanf_match('xx/yy/frame00000000.icpf', 'xx/yy/frame%08d.icpf')
    True
    >>> check_scanf_match('xx/yy//frame00000000.icpf', 'xx/yy/frame%08d.icpf')
    False
    >>> check_scanf_match('xx/yy/frame00000000.icpf', 'xx/yy//frame%08d.icpf')
    False
    >>> check_scanf_match("-1-2+3-4", "%02d%02d%02d%02d")
    True
    >>> check_scanf_match('frame00000240.icpf', 'frame%08d.icpf')
    True
    >>> check_scanf_match('/mnt/hgfs/ZLI-NFLX-10/USCJND/ref/1920x1080/videoSRC001_1920x1080_30.yuv.avi', '/mnt/hgfs/ZLI-NFLX-10/USCJND/ref/1920x1080/videoSRC001_1920x1080_*.yuv.avi')
    True
    """
    try:
        sscanf(string, template)
        return True
    except (FormatError, IncompleteCaptureError):
        pass

    if fnmatch(string, template):
        return True

    return False


def match_any_files(template):
    dir_ = os.path.dirname(template)
    for filename in os.listdir(dir_):
        filepath = dir_ + '/' + filename
        if check_scanf_match(filepath, template):
            return True
    return False


def unroll_dict_of_lists(dict_of_lists):
    """ Unfold a dictionary of lists into a list of dictionaries.

    >>> dict_of_lists = {'norm_type':['normalize'], 'n_estimators':[10, 50], 'random_state': [0]}
    >>> expected = [{'n_estimators': 10, 'norm_type': 'normalize', 'random_state': 0}, {'n_estimators': 50, 'norm_type': 'normalize', 'random_state': 0}]
    >>> unroll_dict_of_lists(dict_of_lists) == expected
    True

    """
    keys = sorted(dict_of_lists.keys()) # normalize order
    list_of_key_value_pairs = []
    for key in keys:
        values = dict_of_lists[key]
        key_value_pairs = []
        for value in values:
            key_value_pairs.append((key, value))
        list_of_key_value_pairs.append(key_value_pairs)

    list_of_key_value_pairs_rearranged = \
        itertools.product(*list_of_key_value_pairs)

    list_of_dicts = []
    for key_value_pairs in list_of_key_value_pairs_rearranged:
        list_of_dicts.append(dict(key_value_pairs))

    return list_of_dicts


def neg_if_even(x):
    """
    >>> neg_if_even(2)
    -1
    >>> neg_if_even(1)
    1
    >>> neg_if_even(0)
    -1
    >>> neg_if_even(-1)
    1
    >>> neg_if_even(-2)
    -1

    """
    return 1 - (x % 2 == 0) * 2


def get_unique_sorted_list(l):
    """
    >>> get_unique_sorted_list([3, 4, 4, 1])
    [1, 3, 4]
    >>> get_unique_sorted_list([])
    []
    """
    return sorted(list(set(l)))


class Timer(object):

    def __enter__(self):
        self.tstart = time()

    def __exit__(self, type, value, traceback):
        print('Elapsed: %s sec' % (time() - self.tstart))


def dedup_value_in_dict(d):
    """
    >>> dedup_value_in_dict({'a': 1, 'b': 1, 'c': 2}) == {'a': 1, 'c': 2}
    True
    """
    reversed_d = dict()
    keys = sorted(d.keys())
    for key in keys:
        value = d[key]
        if value not in reversed_d:
            reversed_d[value] = key
    d_ = dict()
    for value, key in reversed_d.items():
        d_[key] = value
    return d_


class MyTestCase(unittest.TestCase):

    def setUp(self):
        self.verificationErrors = []
        self.maxDiff = None

    def tearDown(self):
        unittest.TestCase.assertEqual(self, [], self.verificationErrors)

    def assertAlmostEqual(self, first, second, places=None, msg=None, delta=None):
        try:
            super().assertAlmostEqual(first, second, places, msg, delta)
        except AssertionError as e:
            self.verificationErrors.append(str(e))

    def assertEqual(self, first, second, msg=None):
        try:
            super().assertEqual(first, second, msg)
        except AssertionError as e:
            self.verificationErrors.append(str(e))

    def assertTrue(self, expr, msg=None):
        try:
            super().assertTrue(expr, msg)
        except AssertionError as e:
            self.verificationErrors.append(str(e))


class QualityRunnerTestMixin(object):

    @staticmethod
    def _run_each_no_assert(runner_class, asset, optional_dict,
                            optional_dict2=None, result_store=None, **more):
        runner = runner_class(
            [asset],
            None,
            fifo_mode=False,
            delete_workdir=True,
            result_store=result_store,
            optional_dict=optional_dict,
            optional_dict2=optional_dict2,
        )
        runner.run(parallelize=False)
        result = runner.results[0]
        return result

    def run_each(self, score, runner_class, asset, optional_dict,
                 optional_dict2=None, result_store=None, places=5, **more):
        result = self._run_each_no_assert(
            runner_class, asset, optional_dict,
            optional_dict2, result_store, **more)
        self.assertAlmostEqual(result[runner_class.get_score_key()],
                               score, places=places)

    def plot_frame_scores(self, ax, *args, **kwargs):
        result = self._run_each_no_assert(*args, **kwargs)
        runner_class, asset, optional_dict = args
        avg_score = result[runner_class.get_score_key()]
        label = [f'avg. {runner_class.TYPE}: {avg_score:.3f}']
        if 'label' in kwargs:
            label += [kwargs['label']]
        label = ', '.join(label)
        ax.set_xlabel('Frame Number')
        ax.set_ylabel('Score')
        ax.plot(result[runner_class.get_scores_key()], label=label)


def find_linear_function_parameters(p1, p2):
    """
    Find parameters of a linear function connecting first_point and second_point

    >>> find_linear_function_parameters((1, 1), (0, 0))
    Traceback (most recent call last):
    ...
    AssertionError: first_point coordinates need to be smaller or equal to second_point coordinates
    >>> find_linear_function_parameters((0, 1), (0, 0))
    Traceback (most recent call last):
    ...
    AssertionError: first_point coordinates need to be smaller or equal to second_point coordinates
    >>> find_linear_function_parameters((1, 0), (0, 0))
    Traceback (most recent call last):
    ...
    AssertionError: first_point coordinates need to be smaller or equal to second_point coordinates
    >>> find_linear_function_parameters((50.0, 30.0), (50.0, 100.0))
    Traceback (most recent call last):
    ...
    AssertionError: first_point and second_point cannot lie on a horizontal or vertical line
    >>> find_linear_function_parameters((50.0, 30.0), (100.0, 30.0))
    Traceback (most recent call last):
    ...
    AssertionError: first_point and second_point cannot lie on a horizontal or vertical line
    >>> find_linear_function_parameters((50.0, 20.0), (110.0, 110.0))
    (1.5, -55.0)
    >>> a, b = find_linear_function_parameters((50.0, 30.0), (110.0, 110.0))
    >>> np.testing.assert_almost_equal(a, 1.333333333333333)
    >>> np.testing.assert_almost_equal(b, -36.666666666666664)
    >>> find_linear_function_parameters((50.0, 30.0), (50.0, 30.0))
    (1, 0)
    >>> find_linear_function_parameters((10.0, 10.0), (50.0, 110.0))
    (2.5, -15.0)

    """
    assert len(p1) == 2, 'first_point needs to have exactly 2 coordinates'
    assert len(p2) == 2, 'second_point needs to have exactly 2 coordinates'
    assert p1[0] <= p2[0] and p1[1] <= p2[1], \
        'first_point coordinates need to be smaller or equal to second_point coordinates'

    if p2[0] - p1[0] == 0 or p2[1] - p1[1] == 0:
        assert p1 == p2, 'first_point and second_point cannot lie on a horizontal or vertical line'
        alpha = 1   # both points are the same
        beta = 0
    else:
        alpha = (p2[1] - p1[1]) / (p2[0] - p1[0])
        beta = p1[1] - (p1[0] * alpha)

    return alpha, beta


def piecewise_linear_mapping(x, knots):
    """
    A piecewise linear mapping function, defined by the boundary points of each segment. For example,
    a function consisting of 3 segments is defined by 4 points. The x-coordinate of each point need to be
    greater that the x-coordinate of the previous point, the y-coordinate needs to be greater or equal.
    The function continues with the same slope for the values below the first point and above the last point.
    INPUT:
        x_in - np.array of values to be mapped
        knots - list of (at least 2) lists with x and y coordinates [[x0, y0], [x1, y1], ...]

    >>> x = np.arange(0.0, 110.0)
    >>> piecewise_linear_mapping(x, [[0, 1], [1, 2], [1, 3]])
    Traceback (most recent call last):
    ...
    AssertionError: The x-coordinate of each point need to be greater that the x-coordinate of the previous point, the y-coordinate needs to be greater or equal.
    >>> piecewise_linear_mapping(x, [[0, 0], []])
    Traceback (most recent call last):
    ...
    AssertionError: Each point needs to have two coordinates [x, y]
    >>> piecewise_linear_mapping(x, [0, 0])
    Traceback (most recent call last):
    ...
    AssertionError: knots needs to be list of lists
    >>> piecewise_linear_mapping(x, [[0, 2], [1, 1]])
    Traceback (most recent call last):
    ...
    AssertionError: The x-coordinate of each point need to be greater that the x-coordinate of the previous point, the y-coordinate needs to be greater or equal.

    >>> knots2160p = [[0.0, -55.0], [95.0, 87.5], [105.0, 105.0], [110.0, 110.0]]
    >>> knots1080p = [[0.0, -36.66], [90.0, 83.04], [95.0, 95.0], [100.0, 100.0]]

    >>> x0 = np.arange(0.0, 95.0, 0.1)
    >>> y0_true = 1.5 * x0 - 55.0
    >>> y0 = piecewise_linear_mapping(x0, knots2160p)
    >>> np.sqrt(np.mean((y0 - y0_true)**2))
    0.0
    >>> x1 = np.arange(0.0, 90.0, 0.1)
    >>> y1_true = 1.33 * x1 - 36.66
    >>> y1 = piecewise_linear_mapping(x1, knots1080p)
    >>> np.sqrt(np.mean((y1 - y1_true) ** 2))
    0.0

    >>> x0 = np.arange(95.0, 105.0, 0.1)
    >>> y0_true = 1.75 * x0 - 78.75
    >>> y0 = piecewise_linear_mapping(x0, knots2160p)
    >>> np.sqrt(np.mean((y0 - y0_true) ** 2))
    0.0
    >>> x1 = np.arange(90.0, 95.0, 0.1)
    >>> y1_true = 2.392 * x1 - 132.24
    >>> y1 = piecewise_linear_mapping(x1, knots1080p)
    >>> np.testing.assert_almost_equal(np.sqrt(np.mean((y1 - y1_true) ** 2)), 0.0)

    >>> x0 = np.arange(105.0, 110.0, 0.1)
    >>> y0 = piecewise_linear_mapping(x0, knots2160p)
    >>> np.sqrt(np.mean((y0 - x0) ** 2))
    0.0
    >>> x1 = np.arange(95.0, 100.0, 0.1)
    >>> y1 = piecewise_linear_mapping(x1, knots1080p)
    >>> np.sqrt(np.mean((y1 - x1) ** 2))
    0.0
    >>> knots_single = [[10.0, 10.0], [50.0, 60.0]]
    >>> x0 = np.arange(0.0, 110.0, 0.1)
    >>> y0 = piecewise_linear_mapping(x0, knots_single)
    >>> y0_true = 1.25 * x0 - 2.5
    >>> np.sqrt(np.mean((y0 - y0_true) ** 2))
    0.0
    """
    assert len(knots) > 1
    n_seg = len(knots) - 1

    y = np.zeros(np.shape(x))

    # construct the function
    for idx in range(n_seg):
        assert isinstance(knots[idx], list) and isinstance(knots[idx + 1], list), \
            'knots needs to be list of lists'
        assert len(knots[idx]) == len(knots[idx + 1]) == 2, \
            'Each point needs to have two coordinates [x, y]'
        assert knots[idx][0] <  knots[idx + 1][0] and \
               knots[idx][1] <= knots[idx + 1][1], \
            'The x-coordinate of each point need to be greater that the x-coordinate of the previous point, ' \
            'the y-coordinate needs to be greater or equal.'

        cond0 = knots[idx][0] <= x
        cond1 = x <= knots[idx + 1][0]

        if knots[idx][1] == knots[idx + 1][1]:  # the segment is horizontal
            y[cond0 & cond1] = knots[idx][1]

            if idx == 0:
                # for points below the defined range
                y[x < knots[idx][0]] = knots[idx][1]
            if idx == n_seg - 1:
                # for points above the defined range
                y[x > knots[idx + 1][0]] = knots[idx][1]

        else:
            slope, offset = find_linear_function_parameters(tuple(knots[idx]),
                                                            tuple(knots[idx + 1]))

            y[cond0 & cond1] = slope * x[cond0 & cond1] + offset

            if idx == 0:
                # for points below the defined range
                y[x < knots[idx][0]] = slope * x[x < knots[idx][0]] + offset
            if idx == n_seg - 1:
                # for points above the defined range
                y[x > knots[idx + 1][0]] = slope * x[x > knots[idx + 1][0]] + offset

    return y


def round_up_to_odd(f):
    """
    >>> round_up_to_odd(32.6)
    33
    >>> round_up_to_odd(33.1)
    35
    """
    return int(np.ceil(f) // 2 * 2 + 1)


class NoPrint(object):

    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, 'w')

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout


def linear_func(x, a, b):
    return a*x + b


def linear_fit(x, y):
    """
    >>> fit = linear_fit([0, 1], [0, 1])
    >>> (fit[0][0], fit[0][1])
    (1.0, 0.0)
    """
    assert isinstance(x, (list, tuple, np.ndarray)), 'x must be a list, tuple, or a numpy array'
    assert len(x) == np.size(x) and len(x) > 0, 'x must be one-dimensional with non-zero length'
    assert isinstance(y, (list, tuple, np.ndarray)), 'y must be a list or a numpy array'
    assert len(y) == np.size(y) and len(y) > 0, 'y must be one-dimensional with non-zero length'
    assert len(x) == len(y), 'x must be the same length as y'

    import scipy.optimize
    return scipy.optimize.curve_fit(linear_func, x, y, [1.0, 0.0])


def map_yuv_type_to_bitdepth(yuv_type):
    """
    >>> map_yuv_type_to_bitdepth('yuv420p')
    8
    >>> map_yuv_type_to_bitdepth('yuv422p')
    8
    >>> map_yuv_type_to_bitdepth('yuv444p')
    8
    >>> map_yuv_type_to_bitdepth('yuv420p10le')
    10
    >>> map_yuv_type_to_bitdepth('yuv422p10le')
    10
    >>> map_yuv_type_to_bitdepth('yuv444p10le')
    10
    >>> map_yuv_type_to_bitdepth('yuv420p12le')
    12
    >>> map_yuv_type_to_bitdepth('yuv422p12le')
    12
    >>> map_yuv_type_to_bitdepth('yuv444p12le')
    12
    >>> map_yuv_type_to_bitdepth('yuv420p16le')
    16
    >>> map_yuv_type_to_bitdepth('yuv422p16le')
    16
    >>> map_yuv_type_to_bitdepth('yuv444p16le')
    16
    >>> map_yuv_type_to_bitdepth('notyuv') is None
    True
    """
    if yuv_type in ['yuv420p', 'yuv422p', 'yuv444p']:
        return 8
    elif yuv_type in ['yuv420p10le', 'yuv422p10le', 'yuv444p10le']:
        return 10
    elif yuv_type in ['yuv420p12le', 'yuv422p12le', 'yuv444p12le']:
        return 12
    elif yuv_type in ['yuv420p16le', 'yuv422p16le', 'yuv444p16le']:
        return 16
    else:
        return None


if __name__ == '__main__':
    import doctest
    doctest.testmod()



import os
import numpy as np
import pandas

from vmaf import plt
from vmaf.core.cross_validation import ModelCrossValidation
from vmaf.core.feature_assembler import FeatureAssembler
from vmaf.core.quality_runner import VmafQualityRunner, \
    VmafQualityRunnerModelMixin
from vmaf.core.result_store import FileSystemResultStore
from vmaf.tools.misc import indices, get_stdout_logger, import_python_file, close_logger, get_file_name_without_extension
from vmaf.config import VmafConfig, DisplayConfig
from vmaf.core.asset import Asset
from vmaf.core.train_test_model import TrainTestModel, RegressorMixin, ClassifierMixin
from vmaf.core.local_explainer import LocalExplainer

__copyright__ = "Copyright 2016-2020, Netflix, Inc."
__license__ = "BSD+Patent"


def read_dataset(dataset, **kwargs):

    groundtruth_key = kwargs['groundtruth_key'] if 'groundtruth_key' in kwargs else None
    skip_asset_with_none_groundtruth = kwargs['skip_asset_with_none_groundtruth'] \
        if 'skip_asset_with_none_groundtruth' in kwargs else False
    content_ids = kwargs['content_ids'] if 'content_ids' in kwargs else None
    asset_ids = kwargs['asset_ids'] if 'asset_ids' in kwargs else None
    workdir_root = kwargs['workdir_root'] if 'workdir_root' in kwargs else VmafConfig.workdir_path()

    # asserts, can add more to the list...
    assert hasattr(dataset, 'dataset_name')
    assert hasattr(dataset, 'ref_videos')
    assert hasattr(dataset, 'dis_videos')

    assert hasattr(dataset, 'yuv_fmt') or all(['yuv_fmt' in ref_video for ref_video in dataset.ref_videos])

    data_set_name = dataset.dataset_name
    ref_videos = dataset.ref_videos
    dis_videos = dataset.dis_videos

    width = dataset.width if hasattr(dataset, 'width') else None
    height = dataset.height if hasattr(dataset, 'height') else None
    yuv_fmt = dataset.yuv_fmt if hasattr(dataset, 'yuv_fmt') else None

    quality_width = dataset.quality_width if hasattr(dataset, 'quality_width') else None
    quality_height = dataset.quality_height if hasattr(dataset, 'quality_height') else None
    resampling_type = dataset.resampling_type if hasattr(dataset, 'resampling_type') else None
    crop_cmd = dataset.crop_cmd if hasattr(dataset, 'crop_cmd') else None
    pad_cmd = dataset.pad_cmd if hasattr(dataset, 'pad_cmd') else None
    workfile_yuv_type = dataset.workfile_yuv_type if hasattr(dataset, 'workfile_yuv_type') else None
    duration_sec = dataset.duration_sec if hasattr(dataset, 'duration_sec') else None
    fps = dataset.fps if hasattr(dataset, 'fps') else None
    start_frame = dataset.start_frame if hasattr(dataset, 'start_frame') else None
    end_frame = dataset.end_frame if hasattr(dataset, 'end_frame') else None

    ref_dict = {}  # dictionary of content_id -> path for ref videos
    for ref_video in ref_videos:
        ref_dict[ref_video['content_id']] = ref_video

    assets = []
    for dis_video in dis_videos:

        if content_ids is not None and dis_video['content_id'] not in content_ids:
            continue

        if asset_ids is not None and dis_video['asset_id'] not in asset_ids:
            continue

        if groundtruth_key is not None:
            groundtruth = dis_video[groundtruth_key]
        else:
            if 'dmos' in dis_video:
                groundtruth = dis_video['dmos']
            elif 'mos' in dis_video:
                groundtruth = dis_video['mos']
            elif 'groundtruth' in dis_video:
                groundtruth = dis_video['groundtruth']
            else:
                groundtruth = None

        if 'os' in dis_video:
            raw_groundtruth = dis_video['os']
        else:
            raw_groundtruth = None

        if 'groundtruth_std' in dis_video:
            groundtruth_std = dis_video['groundtruth_std']
        else:
            groundtruth_std = None

        if 'rebuf_indices' in dis_video:
            rebuf_indices = dis_video['rebuf_indices']
        else:
            rebuf_indices = None

        ref_video = ref_dict[dis_video['content_id']]

        ref_path = ref_video['path']

        ref_yuv_fmt_ = yuv_fmt if yuv_fmt is not None else ref_dict[dis_video['content_id']]['yuv_fmt']
        dis_yuv_fmt_ = dis_video['yuv_fmt'] if 'yuv_fmt' in dis_video else ref_yuv_fmt_

        if width is not None:
            width_ = width
        elif 'width' in ref_video and 'width' not in dis_video:
            width_ = ref_video['width']
        elif 'width' in dis_video and 'width' not in ref_video:
            width_ = dis_video['width']
        elif 'width' in ref_video and 'width' in dis_video:
            assert ref_video['width'] == dis_video['width']
            width_ = ref_video['width']
        else:
            width_ = None

        if height is not None:
            height_ = height
        elif 'height' in ref_video and 'height' not in dis_video:
            height_ = ref_video['height']
        elif 'height' in dis_video and 'height' not in ref_video:
            height_ = dis_video['height']
        elif 'height' in ref_video and 'height' in dis_video:
            assert ref_video['height'] == dis_video['height']
            height_ = ref_video['height']
        else:
            height_ = None

        if quality_width is not None:
            quality_width_ = quality_width
        elif 'quality_width' in dis_video:
            quality_width_ = dis_video['quality_width']
        else:
            quality_width_ = None

        if quality_height is not None:
            quality_height_ = quality_height
        elif 'quality_height' in dis_video:
            quality_height_ = dis_video['quality_height']
        else:
            quality_height_ = None

        if resampling_type is not None:
            resampling_type_ = resampling_type
        elif 'resampling_type' in dis_video:
            resampling_type_ = dis_video['resampling_type']
        else:
            resampling_type_ = None

        if crop_cmd is not None:
            ref_crop_cmd_ = crop_cmd
            dis_crop_cmd_ = crop_cmd
        else:
            if 'crop_cmd' in ref_video:
                ref_crop_cmd_ = ref_video['crop_cmd']
            else:
                ref_crop_cmd_ = None
            if 'crop_cmd' in dis_video:
                dis_crop_cmd_ = dis_video['crop_cmd']
            else:
                dis_crop_cmd_ = None

        if pad_cmd is not None:
            ref_pad_cmd_ = pad_cmd
            dis_pad_cmd_ = pad_cmd
        else:
            if 'pad_cmd' in ref_video:
                ref_pad_cmd_ = ref_video['pad_cmd']
            else:
                ref_pad_cmd_ = None
            if 'pad_cmd' in dis_video:
                dis_pad_cmd_ = dis_video['pad_cmd']
            else:
                dis_pad_cmd_ = None

        if duration_sec is not None:
            duration_sec_ = duration_sec
        elif 'duration_sec' in dis_video:
            duration_sec_ = dis_video['duration_sec']
        else:
            duration_sec_ = None

        if fps is not None:
            fps_ = fps
        elif 'fps' in dis_video:
            fps_ = dis_video['fps']
        else:
            fps_ = None

        if start_frame is not None:
            start_frame_ = start_frame
        elif 'start_frame' in dis_video:
            start_frame_ = dis_video['start_frame']
        else:
            start_frame_ = None

        if end_frame is not None:
            end_frame_ = end_frame
        elif 'end_frame' in dis_video:
            end_frame_ = dis_video['end_frame']
        else:
            end_frame_ = None

        asset_dict = {'ref_yuv_type': ref_yuv_fmt_, 'dis_yuv_type': dis_yuv_fmt_}
        if width_ is not None:
            if asset_dict['ref_yuv_type'] != 'notyuv':
                asset_dict['ref_width'] = width_
            if asset_dict['dis_yuv_type'] != 'notyuv':
                asset_dict['dis_width'] = width_
        if height_ is not None:
            if asset_dict['ref_yuv_type'] != 'notyuv':
                asset_dict['ref_height'] = height_
            if asset_dict['dis_yuv_type'] != 'notyuv':
                asset_dict['dis_height'] = height_
        if groundtruth is not None:
            asset_dict['groundtruth'] = groundtruth
        if raw_groundtruth is not None:
            asset_dict['raw_groundtruth'] = raw_groundtruth
        if groundtruth_std is not None:
            asset_dict['groundtruth_std'] = groundtruth_std
        if quality_width_ is not None:
            asset_dict['quality_width'] = quality_width_
        if quality_height_ is not None:
            asset_dict['quality_height'] = quality_height_
        if resampling_type_ is not None:
            asset_dict['resampling_type'] = resampling_type_

        if ref_crop_cmd_ is not None:
            asset_dict['ref_crop_cmd'] = ref_crop_cmd_
        if dis_crop_cmd_ is not None:
            asset_dict['dis_crop_cmd'] = dis_crop_cmd_

        if ref_pad_cmd_ is not None:
            asset_dict['ref_pad_cmd'] = ref_pad_cmd_
        if dis_pad_cmd_ is not None:
            asset_dict['dis_pad_cmd'] = dis_pad_cmd_

        if duration_sec_ is not None:
            asset_dict['duration_sec'] = duration_sec_
        if workfile_yuv_type is not None:
            asset_dict['workfile_yuv_type'] = workfile_yuv_type
        if rebuf_indices is not None:
            asset_dict['rebuf_indices'] = rebuf_indices
        if fps_ is not None:
            asset_dict['fps'] = fps_
        if start_frame_ is not None:
            asset_dict['start_frame'] = start_frame_
        if end_frame_ is not None:
            asset_dict['end_frame'] = end_frame_

        if 'ref_start_frame' in ref_video:
            asset_dict['ref_start_frame'] = ref_video['ref_start_frame']
        if 'dis_start_frame' in dis_video:
            asset_dict['dis_start_frame'] = dis_video['dis_start_frame']
        if 'ref_end_frame' in ref_video:
            asset_dict['ref_end_frame'] = ref_video['ref_end_frame']
        if 'dis_end_frame' in dis_video:
            asset_dict['dis_end_frame'] = dis_video['dis_end_frame']

        if 'enc_width' in dis_video:
            asset_dict['dis_enc_width'] = dis_video['enc_width']
        if 'enc_height' in dis_video:
            asset_dict['dis_enc_height'] = dis_video['enc_height']
        if 'enc_bitdepth' in dis_video:
            asset_dict['dis_enc_bitdepth'] = dis_video['enc_bitdepth']

        if groundtruth is None and skip_asset_with_none_groundtruth:
            pass
        else:
            asset = Asset(dataset=data_set_name,
                          content_id=dis_video['content_id'],
                          asset_id=dis_video['asset_id'],
                          workdir_root=workdir_root,
                          ref_path=ref_path,
                          dis_path=dis_video['path'],
                          asset_dict=asset_dict,
                          )
            assets.append(asset)

    return assets


def compare_two_quality_runners_on_dataset(
        test_dataset, first_runner_class, second_runner_class,
        result_store,
        parallelize=True, fifo_mode=True,
        aggregate_method=np.mean,
        type='regressor',
        num_resample=1000,
        seed_resample=None,
        ax_plcc=None,
        ax_srocc=None,
        **kwargs):

    def _get_stat(df: pandas.DataFrame,
                  xcol: str,
                  ycol: str,
                  ) -> dict:
        plcc = df[ycol].corr(df[xcol], method='pearson')
        srocc = df[ycol].corr(df[xcol], method='spearman')
        return {'plcc': plcc, 'srocc': srocc}

    first_test_assets, first_results = run_test_on_dataset(
        test_dataset, first_runner_class, None, result_store, None,
        parallelize, fifo_mode, aggregate_method, type, **kwargs)

    second_test_assets, second_results = run_test_on_dataset(
        test_dataset, second_runner_class, None, result_store, None,
        parallelize, fifo_mode, aggregate_method, type, **kwargs)

    # collect data to list of dictionaries
    ds = list()
    assert len(first_test_assets) == len(second_test_assets) == len(first_results) == len(second_results)
    for first_test_asset, first_result, second_test_asset, second_result in zip(first_test_assets, first_results, second_test_assets, second_results):
        assert first_test_asset.groundtruth is not None
        assert second_test_asset.groundtruth is not None
        assert first_test_asset.groundtruth == second_test_asset.groundtruth
        d = {
            'groundtruth': first_test_asset.groundtruth,
            'first_prediction': first_result[first_runner_class.get_score_key()],
            'second_prediction': second_result[second_runner_class.get_score_key()],
        }
        ds.append(d)
    df = pandas.DataFrame(ds)

    # bootstrapping
    np.random.seed(seed_resample)
    xs = list()
    ys = list()
    xs2 = list()
    ys2 = list()
    for _ in range(num_resample):
        dfb = df.sample(n=df.shape[0], replace=True)
        d_stat_first = _get_stat(dfb, 'groundtruth', 'first_prediction')
        d_stat_second = _get_stat(dfb, 'groundtruth', 'second_prediction')
        x = d_stat_first['plcc']
        y = d_stat_second['plcc']
        x2 = d_stat_first['srocc']
        y2 = d_stat_second['srocc']
        xs.append(x)
        ys.append(y)
        xs2.append(x2)
        ys2.append(y2)

    ci95_xs = [np.percentile(xs, 2.5), np.percentile(xs, 97.5)]
    ci95_ys = [np.percentile(ys, 2.5), np.percentile(ys, 97.5)]
    diffs = np.array(ys) - np.array(xs)
    ci95_diffs = [np.percentile(diffs, 2.5), np.percentile(diffs, 97.5)]
    if ax_plcc is not None:
        ax_plcc.scatter(xs, ys, alpha=0.2, label='PLCC with resampling')
        ax_plcc.plot([min(xs), max(xs)], [min(xs), max(xs)], '-r')
        ax_plcc.set_xlabel(f'{first_runner_class.TYPE} 95%-CI: [{ci95_xs[0]:.4f}, {ci95_xs[1]:.4f}]')
        ax_plcc.set_ylabel(f'{second_runner_class.TYPE} 95%-CI: [{ci95_ys[0]:.4f}, {ci95_ys[1]:.4f}]')
        ax_plcc.set_title(f'({second_runner_class.TYPE} - {first_runner_class.TYPE}) 95%-CI: [{ci95_diffs[0]:.4f}, {ci95_diffs[1]:.4f}]')
        ax_plcc.grid()
        ax_plcc.legend()

    ci95_xs2 = [np.percentile(xs2, 2.5), np.percentile(xs2, 97.5)]
    ci95_ys2 = [np.percentile(ys2, 2.5), np.percentile(ys2, 97.5)]
    diffs2 = np.array(ys2) - np.array(xs2)
    ci95_diffs2 = [np.percentile(diffs2, 2.5), np.percentile(diffs2, 97.5)]
    if ax_srocc is not None:
        ax_srocc.scatter(xs2, ys2, alpha=0.2, label='SROCC with resampling')
        ax_srocc.plot([min(xs2), max(xs2)], [min(xs2), max(xs2)], '-r')
        ax_srocc.set_xlabel(f'{first_runner_class.TYPE} 95%-CI: [{ci95_xs2[0]:.4f}, {ci95_xs2[1]:.4f}]')
        ax_srocc.set_ylabel(f'{second_runner_class.TYPE} 95%-CI: [{ci95_ys2[0]:.4f}, {ci95_ys2[1]:.4f}]')
        ax_srocc.set_title(f'({second_runner_class.TYPE} - {first_runner_class.TYPE})  95%-CI: [{ci95_diffs2[0]:.4f}, {ci95_diffs2[1]:.4f}]')
        ax_srocc.grid()
        ax_srocc.legend()

    return {
        'plcc': list(zip(xs, ys)),
        'srocc': list(zip(xs2, ys2)),
        'plcc_ci95_first': ci95_xs,
        'plcc_ci95_second': ci95_ys,
        'plcc_ci95_diff': ci95_diffs,
        'srocc_ci95_first': ci95_xs2,
        'srocc_ci95_second': ci95_ys2,
        'srocc_ci95_diff': ci95_diffs2,
    }


def run_test_on_dataset(test_dataset, runner_class, ax,
                    result_store, model_filepath,
                    parallelize=True, fifo_mode=True,
                    aggregate_method=np.mean,
                    type='regressor',
                    **kwargs):

    test_assets = read_dataset(test_dataset, **kwargs)
    test_raw_assets = None
    try:
        for test_asset in test_assets:
            assert test_asset.groundtruth is not None
    except AssertionError:
        # no groundtruth, try to do subjective modeling
        from sureal.dataset_reader import RawDatasetReader
        from sureal.subjective_model import DmosModel
        subj_model_class = kwargs['subj_model_class'] if 'subj_model_class' in kwargs and kwargs['subj_model_class'] is not None else DmosModel
        dataset_reader_class = kwargs['dataset_reader_class'] if 'dataset_reader_class' in kwargs else RawDatasetReader
        subjective_model = subj_model_class(dataset_reader_class(test_dataset))
        subjective_model.run_modeling(**kwargs)
        test_dataset_aggregate = subjective_model.to_aggregated_dataset(**kwargs)
        test_raw_assets = test_assets
        test_assets = read_dataset(test_dataset_aggregate, **kwargs)

    optional_dict = kwargs['optional_dict'] if 'optional_dict' in kwargs else None

    if model_filepath is not None:
        if not optional_dict:
            optional_dict = {}
        optional_dict['model_filepath'] = model_filepath
        if 'model_720_filepath' in kwargs and kwargs['model_720_filepath'] is not None:
            optional_dict['720model_filepath'] = kwargs['model_720_filepath']
        if 'model_480_filepath' in kwargs and kwargs['model_480_filepath'] is not None:
            optional_dict['480model_filepath'] = kwargs['model_480_filepath']
        if 'model_2160_filepath' in kwargs and kwargs['model_2160_filepath'] is not None:
            optional_dict['2160model_filepath'] = kwargs['model_2160_filepath']

    if 'enable_transform_score' in kwargs and kwargs['enable_transform_score'] is not None:
        if not optional_dict:
            optional_dict = {}
        optional_dict['enable_transform_score'] = kwargs['enable_transform_score']

    if 'disable_clip_score' in kwargs and kwargs['disable_clip_score'] is not None:
        if not optional_dict:
            optional_dict = {}
        optional_dict['disable_clip_score'] = kwargs['disable_clip_score']

    if 'subsample' in kwargs and kwargs['subsample'] is not None:
        if not optional_dict:
            optional_dict = {}
        optional_dict['subsample'] = kwargs['subsample']

    if 'additional_optional_dict' in kwargs and kwargs['additional_optional_dict'] is not None:
        assert isinstance(kwargs['additional_optional_dict'], dict)
        if not optional_dict:
            optional_dict = {}
        optional_dict.update(kwargs['additional_optional_dict'])

    if 'processes' in kwargs and kwargs['processes'] is not None:
        assert isinstance(kwargs['processes'], int)
        processes = kwargs['processes']
    else:
        processes = None
    if processes is not None:
        assert parallelize is True, 'if processes is not None, parallelize must be True'

    # run
    runner = runner_class(
        test_assets,
        None, fifo_mode=fifo_mode,
        delete_workdir=True,
        result_store=result_store,
        optional_dict=optional_dict,
        optional_dict2=None,
    )
    runner.run(parallelize=parallelize, processes=processes)
    results = runner.results

    for result in results:
        result.set_score_aggregate_method(aggregate_method)

    try:
        model_type = runner.get_train_test_model_class()
    except:
        if type == 'regressor':
            model_type = RegressorMixin
        elif type == 'classifier':
            model_type = ClassifierMixin
        else:
            assert False

    split_test_indices_for_perf_ci = kwargs['split_test_indices_for_perf_ci'] \
        if 'split_test_indices_for_perf_ci' in kwargs else False

    # plot
    groundtruths = list(map(lambda asset: asset.groundtruth, test_assets))
    predictions = list(map(lambda result: result[runner_class.get_score_key()], results))
    raw_grountruths = None if test_raw_assets is None else \
        list(map(lambda asset: asset.raw_groundtruth, test_raw_assets))
    groundtruths_std = None if test_assets is None else \
        list(map(lambda asset: asset.groundtruth_std, test_assets))
    try:
        predictions_bagging = list(map(lambda result: result[runner_class.get_bagging_score_key()], results))
        predictions_stddev = list(map(lambda result: result[runner_class.get_stddev_score_key()], results))
        predictions_ci95_low = list(map(lambda result: result[runner_class.get_ci95_low_score_key()], results))
        predictions_ci95_high = list(map(lambda result: result[runner_class.get_ci95_high_score_key()], results))
        predictions_all_models = list(map(lambda result: result[runner_class.get_all_models_score_key()], results))

        # need to revert the list of lists, so that the outer list has the predictions for each model separately
        predictions_all_models = np.array(predictions_all_models).T.tolist()
        num_models = np.shape(predictions_all_models)[0]

        stats = model_type.get_stats(groundtruths, predictions,
                                     ys_label_raw=raw_grountruths,
                                     ys_label_pred_bagging=predictions_bagging,
                                     ys_label_pred_stddev=predictions_stddev,
                                     ys_label_pred_ci95_low=predictions_ci95_low,
                                     ys_label_pred_ci95_high=predictions_ci95_high,
                                     ys_label_pred_all_models=predictions_all_models,
                                     ys_label_stddev=groundtruths_std,
                                     split_test_indices_for_perf_ci=split_test_indices_for_perf_ci)
    except Exception as e:
        print('Warning: stats calculation failed, fall back to default stats calculation: {}'.format(e))
        stats = model_type.get_stats(groundtruths, predictions,
                                     ys_label_raw=raw_grountruths,
                                     ys_label_stddev=groundtruths_std,
                                     split_test_indices_for_perf_ci=split_test_indices_for_perf_ci)
        num_models = 1

    print('Stats on testing data: {}'.format(model_type.format_stats_for_print(stats)))

    # printing stats if multiple models are present
    if 'SRCC_across_model_distribution' in stats \
            and 'PCC_across_model_distribution' in stats \
            and 'RMSE_across_model_distribution' in stats:
        print('Stats on testing data (across multiple models, using all test indices): {}'.format(
            model_type.format_across_model_stats_for_print(model_type.extract_across_model_stats(stats))))

    if split_test_indices_for_perf_ci:
        print('Stats on testing data (single model, multiple test sets): {}'
              .format(model_type.format_stats_across_test_splits_for_print(model_type.extract_across_test_splits_stats(stats))))

    if ax is not None:
        content_ids = list(map(lambda asset: asset.content_id, test_assets))

        if 'point_label' in kwargs and kwargs['point_label'] is not None:
            if kwargs['point_label'] == 'asset_id':
                point_labels = list(map(lambda asset: asset.asset_id, test_assets))
            elif kwargs['point_label'] == 'dis_path':
                point_labels = list(map(lambda asset: get_file_name_without_extension(asset.dis_path), test_assets))
            else:
                raise AssertionError("Unknown point_label {}".format(kwargs['point_label']))
        else:
            point_labels = None

        model_type.plot_scatter(ax, stats, content_ids=content_ids, point_labels=point_labels, **kwargs)
        ax.set_xlabel('True Score')
        ax.set_ylabel("Predicted Score")
        ax.grid()
        ax.set_title("{runner}{num_models}\n{stats}".format(
            dataset=test_assets[0].dataset,
            runner=runner_class.TYPE,
            stats=model_type.format_stats_for_plot(stats),
            num_models=", {} models".format(num_models) if num_models > 1 else "",
        ))

    return test_assets, results


def print_matplotlib_warning():
    print("Warning: cannot import matplotlib, no picture displayed. "
          "If you are on Mac OS and have installed matplotlib, you "
          "possibly need to run: \nsudo pip uninstall python-dateutil \n"
          "sudo pip install python-dateutil==2.2 \n"
          "Refer to: https://stackoverflow.com/questions/27630114/matplotlib-issue-on-os-x-importerror-cannot-import-name-thread")


def train_test_vmaf_on_dataset(train_dataset, test_dataset,
                               feature_param, model_param,
                               train_ax, test_ax, result_store,
                               logger=None, fifo_mode=True,
                               output_model_filepath=None,
                               aggregate_method=np.mean,
                               **kwargs):

    train_assets = read_dataset(train_dataset, **kwargs)
    train_raw_assets = None
    try:
        for train_asset in train_assets:
            assert train_asset.groundtruth is not None
    except AssertionError:
        # no groundtruth, try to do subjective modeling
        from sureal.dataset_reader import RawDatasetReader
        from sureal.subjective_model import DmosModel
        subj_model_class = kwargs['subj_model_class'] if 'subj_model_class' in kwargs and kwargs['subj_model_class'] is not None else DmosModel
        dataset_reader_class = kwargs['dataset_reader_class'] if 'dataset_reader_class' in kwargs else RawDatasetReader
        subjective_model = subj_model_class(dataset_reader_class(train_dataset))
        subjective_model.run_modeling(**kwargs)
        train_dataset_aggregate = subjective_model.to_aggregated_dataset(**kwargs)
        train_raw_assets = train_assets
        train_assets = read_dataset(train_dataset_aggregate, **kwargs)

    parallelize = kwargs['parallelize'] if 'parallelize' in kwargs else True
    isinstance(parallelize, bool)

    processes = kwargs['processes'] if 'processes' in kwargs else None
    if processes is not None:
        assert isinstance(processes, int) and processes > 0
    if processes is not None:
        assert parallelize is True, 'if processes is not None, parallelize must be True'

    assert hasattr(feature_param, 'feature_dict')
    feature_dict = feature_param.feature_dict
    feature_option_dict = feature_param.feature_optional_dict if hasattr(feature_param, 'feature_optional_dict') else None

    train_fassembler = FeatureAssembler(
        feature_dict=feature_dict,
        feature_option_dict=feature_option_dict,
        assets=train_assets,
        logger=logger,
        fifo_mode=fifo_mode,
        delete_workdir=True,
        result_store=result_store,
        optional_dict=None,  # WARNING: feature param not passed
        optional_dict2=None,
        parallelize=parallelize,
        processes=processes,
    )
    train_fassembler.run()
    train_features = train_fassembler.results

    for result in train_features:
        result.set_score_aggregate_method(aggregate_method)

    model_type = model_param.model_type
    model_param_dict = model_param.model_param_dict

    model_class = TrainTestModel.find_subclass(model_type)

    train_xys = model_class.get_xys_from_results(train_features)
    train_xs = model_class.get_xs_from_results(train_features)
    train_ys = model_class.get_ys_from_results(train_features)

    model = model_class(model_param_dict, logger)

    model.train(train_xys, feature_option_dict=feature_option_dict, **kwargs)

    # append additional information to model before saving, so that
    # VmafQualityRunner can read and process
    model.append_info('feature_dict', feature_param.feature_dict)  # need feature_dict so that VmafQualityRunner knows how to call FeatureAssembler
    if 'score_clip' in model_param_dict:
        VmafQualityRunner.set_clip_score(model, model_param_dict['score_clip'])
    if 'score_transform' in model_param_dict:
        VmafQualityRunner.set_transform_score(model, model_param_dict['score_transform'])

    train_ys_pred = VmafQualityRunner.predict_with_model(model, train_xs, **kwargs)['ys_pred']

    raw_groundtruths = None if train_raw_assets is None else \
        list(map(lambda asset: asset.raw_groundtruth, train_raw_assets))

    train_stats = model.get_stats(train_ys['label'], train_ys_pred, ys_label_raw=raw_groundtruths)

    log = 'Stats on training data: {}'.format(model.format_stats_for_print(train_stats))
    if logger:
        logger.info(log)
    else:
        print(log)

    # save model
    if output_model_filepath is not None:
        format = os.path.splitext(output_model_filepath)[1]
        supported_formats = ['.pkl', '.json']
        VmafQualityRunnerModelMixin._assert_extension_format(supported_formats, format)
        if '.pkl' in format:
            model.to_file(output_model_filepath, format='pkl')
        elif '.json' in format:
            model.to_file(output_model_filepath, format='json', combined=True)
        else:
            assert False

    if train_ax is not None:
        train_content_ids = list(map(lambda asset: asset.content_id, train_assets))
        model_class.plot_scatter(train_ax, train_stats, content_ids=train_content_ids)

        train_ax.set_xlabel('True Score')
        train_ax.set_ylabel("Predicted Score")
        train_ax.grid()
        train_ax.set_title("Dataset: {dataset}, Model: {model}\n{stats}".format(
            dataset=train_dataset.dataset_name,
            model=model.model_id,
            stats=model_class.format_stats_for_plot(train_stats)
        ))

    # === test model on test dataset ===

    if test_dataset is None:
        test_assets = None
        test_stats = None
        test_fassembler = None
    else:
        test_assets = read_dataset(test_dataset, **kwargs)
        test_raw_assets = None
        try:
            for test_asset in test_assets:
                assert test_asset.groundtruth is not None
        except AssertionError:
            # no groundtruth, try to do subjective modeling
            from sureal.dataset_reader import RawDatasetReader
            from sureal.subjective_model import DmosModel
            subj_model_class = kwargs['subj_model_class'] if 'subj_model_class' in kwargs and kwargs['subj_model_class'] is not None else DmosModel
            dataset_reader_class = kwargs['dataset_reader_class'] if 'dataset_reader_class' in kwargs else RawDatasetReader
            subjective_model = subj_model_class(dataset_reader_class(test_dataset))
            subjective_model.run_modeling(**kwargs)
            test_dataset_aggregate = subjective_model.to_aggregated_dataset(**kwargs)
            test_raw_assets = test_assets
            test_assets = read_dataset(test_dataset_aggregate, **kwargs)

        test_fassembler = FeatureAssembler(
            feature_dict=feature_dict,
            feature_option_dict=feature_option_dict,
            assets=test_assets,
            logger=logger,
            fifo_mode=fifo_mode,
            delete_workdir=True,
            result_store=result_store,
            optional_dict=None,  # WARNING: feature param not passed
            optional_dict2=None,
            parallelize=parallelize,
        )
        test_fassembler.run()
        test_features = test_fassembler.results

        for result in test_features:
            result.set_score_aggregate_method(aggregate_method)

        test_xs = model_class.get_xs_from_results(test_features)
        test_ys = model_class.get_ys_from_results(test_features)

        test_ys_pred = VmafQualityRunner.predict_with_model(model, test_xs, **kwargs)['ys_pred']

        raw_groundtruths = None if test_raw_assets is None else \
            list(map(lambda asset: asset.raw_groundtruth, test_raw_assets))

        test_stats = model.get_stats(test_ys['label'], test_ys_pred, ys_label_raw=raw_groundtruths)

        log = 'Stats on testing data: {}'.format(model_class.format_stats_for_print(test_stats))
        if logger:
            logger.info(log)
        else:
            print(log)

        if test_ax is not None:
            test_content_ids = list(map(lambda asset: asset.content_id, test_assets))
            model_class.plot_scatter(test_ax, test_stats, content_ids=test_content_ids)
            test_ax.set_xlabel('True Score')
            test_ax.set_ylabel("Predicted Score")
            test_ax.grid()
            test_ax.set_title("Dataset: {dataset}, Model: {model}\n{stats}".format(
                dataset=test_dataset.dataset_name,
                model=model.model_id,
                stats=model_class.format_stats_for_plot(test_stats)
            ))

    return train_fassembler, train_assets, train_stats, test_fassembler, test_assets, test_stats, model


def construct_kfold_list(assets, contentid_groups):
    # construct cross validation kfold input list
    content_ids = list(map(lambda asset: asset.content_id, assets))
    kfold = []
    for curr_content_group in contentid_groups:
        curr_indices = indices(content_ids, lambda x: x in curr_content_group)
        kfold.append(curr_indices)
    return kfold


def cv_on_dataset(dataset, feature_param, model_param, ax, result_store,
                  contentid_groups, logger=None, aggregate_method=np.mean):

    assets = read_dataset(dataset)
    kfold = construct_kfold_list(assets, contentid_groups)

    fassembler = FeatureAssembler(
        feature_dict=feature_param.feature_dict,
        feature_option_dict=None,  # FIXME: as set to None, potential bug with inconsistent behavior with VmafQualityRunner
        assets=assets,
        logger=logger,
        delete_workdir=True,
        result_store=result_store,
        optional_dict=None,  # WARNING: feature param not passed
        optional_dict2=None,
        parallelize=True, fifo_mode=True,
        # parallelize=False, fifo_mode=False, # VQM
    )
    fassembler.run()
    results = fassembler.results

    for result in results:
        result.set_score_aggregate_method(aggregate_method)

    model_class = TrainTestModel.find_subclass(model_param.model_type)
    # run nested kfold cv for each combintation
    cv_output = ModelCrossValidation.run_kfold_cross_validation(
        model_class,
        model_param.model_param_dict,
        results,
        kfold,
        logger=logger,
    )

    print('Feature parameters: {}'.format(feature_param.feature_dict))
    print('Model type: {}'.format(model_param.model_type))
    print('Model parameters: {}'.format(model_param.model_param_dict))
    print('Stats: {}'.format(model_class.format_stats_for_print(cv_output['aggr_stats'])))

    if ax is not None:
        model_class.plot_scatter(ax, cv_output['aggr_stats'], content_ids=cv_output['contentids'])
        ax.set_xlabel('True Score')
        ax.set_ylabel("Predicted Score")
        ax.grid()
        ax.set_title("Dataset: {dataset}, Model: {model},\n{stats}".format(
            dataset=dataset.dataset_name,
            model=model_param.model_type,
            stats=model_class.format_stats_for_plot(cv_output['aggr_stats'])
        ))

    return assets, cv_output


def run_remove_results_for_dataset(result_store, dataset, executor_class):
    assets = read_dataset(dataset)
    executor = executor_class(assets=assets, logger=None, result_store=result_store)
    executor.remove_results()


def run_vmaf_cv(train_dataset_filepath,
                test_dataset_filepath,
                param_filepath,
                output_model_filepath=None,
                **kwargs):

    result_store_dir = kwargs['result_store_dir'] if 'result_store_dir' in kwargs else VmafConfig.file_result_store_path()

    parallelize = kwargs['parallelize'] if 'parallelize' in kwargs else True
    isinstance(parallelize, bool)

    logger = get_stdout_logger()
    result_store = FileSystemResultStore(result_store_dir)

    train_dataset = import_python_file(train_dataset_filepath)
    test_dataset = import_python_file(test_dataset_filepath) if test_dataset_filepath is not None else None

    param = import_python_file(param_filepath)

    # === plot scatter ===

    nrows = 1
    ncols = 2
    fig, axs = plt.subplots(figsize=(5*ncols, 5*nrows), nrows=nrows, ncols=ncols)

    train_test_vmaf_on_dataset(train_dataset, test_dataset, param, param, axs[0], axs[1],
                               result_store, logger=None,
                               output_model_filepath=output_model_filepath,
                               **kwargs)

    if 'xlim' in kwargs:
        axs[0].set_xlim(kwargs['xlim'])
        axs[1].set_xlim(kwargs['xlim'])

    if 'ylim' in kwargs:
        axs[0].set_ylim(kwargs['ylim'])
        axs[1].set_ylim(kwargs['ylim'])

    bbox = {'facecolor':'white', 'alpha':1, 'pad':20}
    axs[0].annotate('Training Set', xy=(0.1, 0.85), xycoords='axes fraction', bbox=bbox)
    axs[1].annotate('Testing Set', xy=(0.1, 0.85), xycoords='axes fraction', bbox=bbox)

    plt.tight_layout()

    # === clean up ===
    close_logger(logger)


def run_vmaf_kfold_cv(dataset_filepath,
                      contentid_groups,
                      param_filepath,
                      aggregate_method,
                      result_store_dir=VmafConfig.file_result_store_path(),
                      ):

    logger = get_stdout_logger()
    result_store = FileSystemResultStore(result_store_dir)
    dataset = import_python_file(dataset_filepath)
    param = import_python_file(param_filepath)

    fig, ax = plt.subplots(figsize=(5, 5), nrows=1, ncols=1)

    cv_on_dataset(dataset, param, param, ax, result_store, contentid_groups,
                  logger, aggregate_method)

    ax.set_xlim([0, 120])
    ax.set_ylim([0, 120])
    plt.tight_layout()

    # === clean up ===
    close_logger(logger)


def explain_model_on_dataset(model, test_assets_selected_indexs,
                             test_dataset_filepath,
                             result_store_dir=VmafConfig.file_result_store_path()):

    def print_assets(test_assets):
        print('\n'.join(map(
            lambda tasset: "Asset {i}: {name}".format(
                i=tasset[0], name=get_file_name_without_extension(tasset[1].dis_path)),
            enumerate(test_assets)
        )))

    test_dataset = import_python_file(test_dataset_filepath)
    test_assets = read_dataset(test_dataset)
    print_assets(test_assets)
    print("Assets selected for local explanation: {}".format(
        test_assets_selected_indexs))
    result_store = FileSystemResultStore(result_store_dir)
    test_assets = [test_assets[i] for i in test_assets_selected_indexs]
    test_fassembler = FeatureAssembler(
        feature_dict=model.model_dict['feature_dict'],
        feature_option_dict=None,  # FIXME: as set to None, potential bug with inconsistent behavior with VmafQualityRunner
        assets=test_assets,
        logger=None,
        fifo_mode=True,
        delete_workdir=True,
        result_store=result_store,
        optional_dict=None,  # WARNING: feature param not passed
        optional_dict2=None,
        parallelize=True,
    )
    test_fassembler.run()
    test_feature_results = test_fassembler.results
    test_xs = model.get_xs_from_results(test_feature_results)
    test_ys = model.get_ys_from_results(test_feature_results)
    test_ys_pred = model.predict(test_xs)['ys_label_pred']
    explainer = LocalExplainer(neighbor_samples=1000)
    test_exps = explainer.explain(model, test_xs)

    explainer.print_explanations(test_exps, assets=test_assets, ys=test_ys, ys_pred=test_ys_pred)
    explainer.plot_explanations(test_exps, assets=test_assets, ys=test_ys, ys_pred=test_ys_pred)
    DisplayConfig.show()


def generate_dataset_from_raw(raw_dataset_filepath, output_dataset_filepath, **kwargs):
    if raw_dataset_filepath:
        from sureal.subjective_model import DmosModel
        subj_model_class = kwargs['subj_model_class'] if 'subj_model_class' in kwargs else DmosModel
        content_ids = kwargs['content_ids'] if 'content_ids' in kwargs else None
        asset_ids = kwargs['asset_ids'] if 'asset_ids' in kwargs else None
        subjective_model = subj_model_class.from_dataset_file(raw_dataset_filepath,
                                                              content_ids=content_ids,
                                                              asset_ids=asset_ids)
        subjective_model.run_modeling(**kwargs)
        subjective_model.to_aggregated_dataset_file(output_dataset_filepath, **kwargs)


def run_vmaf_cv_from_raw(train_dataset_raw_filepath, test_dataset_raw_filepath,
                    param_filepath, output_model_filepath, **kwargs):
    if 'train_quality_wh' in kwargs and kwargs['train_quality_wh'] is not None:
        train_quality_width, train_quality_height = kwargs['train_quality_wh']
    else:
        train_quality_width = None
        train_quality_height = None

    if 'test_quality_wh' in kwargs and kwargs['test_quality_wh'] is not None:
        test_quality_width, test_quality_height = kwargs['test_quality_wh']
    else:
        test_quality_width = None
        test_quality_height = None

    if 'train_transform_final' in kwargs and kwargs['train_transform_final'] is not None:
        train_transform_final = kwargs['train_transform_final']
    else:
        train_transform_final = None

    if 'test_transform_final' in kwargs and kwargs['test_transform_final'] is not None:
        test_transform_final = kwargs['test_transform_final']
    else:
        test_transform_final = None

    workspace_path = kwargs['workspace_path'] if 'workspace_path' in kwargs else VmafConfig.workspace_path()

    train_output_dataset_filepath = os.path.join(workspace_path, 'dataset', 'train_dataset.py')
    generate_dataset_from_raw(raw_dataset_filepath=train_dataset_raw_filepath,
                     output_dataset_filepath=train_output_dataset_filepath,
                     quality_width=train_quality_width,
                     quality_height=train_quality_height,
                     transform_final=train_transform_final,
                     **kwargs)

    test_output_dataset_filepath = os.path.join(workspace_path, 'dataset', 'test_dataset.py') \
        if test_dataset_raw_filepath is not None else None
    generate_dataset_from_raw(raw_dataset_filepath=test_dataset_raw_filepath,
                     output_dataset_filepath=test_output_dataset_filepath,
                     quality_width=test_quality_width,
                     quality_height=test_quality_height,
                     transform_final=test_transform_final,
                     **kwargs)

    run_vmaf_cv(
        train_dataset_filepath=train_output_dataset_filepath,
        test_dataset_filepath=test_output_dataset_filepath,
        param_filepath=param_filepath,
        output_model_filepath=output_model_filepath,
        **kwargs
    )



dataset_name = 'example'

quality_width = 1920
quality_height = 1080

from vmaf.config import VmafConfig

ref_videos = [
    {'content_id': 0, 'path': VmafConfig.resource_path('yuv', 'checkerboard_1920_1080_10_3_0_0.yuv'), 'yuv_fmt': 'yuv420p', 'width': 1920, 'height': 1080},

    {'content_id': 1, 'path': VmafConfig.resource_path('yuv', 'flat_1920_1080_0.yuv'), 'yuv_fmt': 'yuv420p', 'width': 720, 'height': 480},
]

dis_videos = [
    {'content_id': 0, 'asset_id': 0, 'dmos': 100, 'path': VmafConfig.resource_path('yuv', 'checkerboard_1920_1080_10_3_0_0.yuv'), 'yuv_fmt': 'yuv420p', 'width': 1920, 'height': 1080}, # ref
    {'content_id': 0, 'asset_id': 1, 'dmos': 50,  'path': VmafConfig.resource_path('yuv', 'checkerboard_1920_1080_10_3_1_0.264'), 'yuv_fmt': 'notyuv',},

    {'content_id': 1, 'asset_id': 2, 'dmos': 100,  'path': VmafConfig.resource_path('yuv', 'flat_1920_1080_0.yuv'), 'yuv_fmt': 'yuv420p', 'width': 720, 'height': 480}, # ref
    {'content_id': 1, 'asset_id': 3, 'dmos': 80,  'path': VmafConfig.resource_path('yuv', 'flat_1920_1080_10.264'), 'yuv_fmt': 'notyuv',},
]



from __future__ import absolute_import

__copyright__ = "Copyright 2016-2020, Netflix, Inc."
__license__ = "BSD+Patent"

import unittest

from vmaf.core.feature_assembler import FeatureAssembler
from vmaf.core.feature_extractor import VmafFeatureExtractor, FeatureExtractor, \
    MomentFeatureExtractor

from test.testutil import set_default_576_324_videos_for_testing


class FeatureAssemblerTest(unittest.TestCase):

    def tearDown(self):
        if hasattr(self, 'fassembler'):
            self.fassembler.remove_results()
        pass

    def test_get_fextractor_subclasses(self):
        fextractor_subclasses = FeatureExtractor.get_subclasses_recursively()
        self.assertTrue(VmafFeatureExtractor in fextractor_subclasses)
        self.assertTrue(MomentFeatureExtractor in fextractor_subclasses)

    def test_feature_assembler_whole_feature(self):

        ref_path, dis_path, asset, asset_original = set_default_576_324_videos_for_testing()

        self.fassembler = FeatureAssembler(
            feature_dict={'VMAF_feature': 'all'},
            feature_option_dict=None,
            assets=[asset, asset_original],
            logger=None,
            fifo_mode=True,
            delete_workdir=True,
            result_store=None,
            optional_dict=None,
            optional_dict2=None,
            parallelize=True,
            processes=None,
        )

        self.fassembler.run()

        results = self.fassembler.results

        self.assertAlmostEqual(results[0]['VMAF_feature_vif_score'], 0.44609306249999997, places=4)
        self.assertAlmostEqual(results[0]['VMAF_feature_motion_score'], 4.0498253541666669, places=4)
        self.assertAlmostEqual(results[0]['VMAF_feature_adm2_score'], 0.9345149030293786, places=4)
        self.assertAlmostEqual(results[0]['VMAF_feature_ansnr_score'], 23.509571520833333, places=4)

        self.assertAlmostEqual(results[1]['VMAF_feature_vif_score'], 1.0, places=4)
        self.assertAlmostEqual(results[1]['VMAF_feature_motion_score'], 4.0498253541666669, places=4)
        self.assertAlmostEqual(results[1]['VMAF_feature_adm2_score'], 1.0, places=4)
        self.assertAlmostEqual(results[1]['VMAF_feature_ansnr_score'], 31.271439270833337, places=4)

    def test_feature_assembler_whole_feature_processes(self):

        ref_path, dis_path, asset, asset_original = set_default_576_324_videos_for_testing()

        self.fassembler = FeatureAssembler(
            feature_dict={'VMAF_feature': 'all'},
            feature_option_dict=None,
            assets=[asset, asset_original],
            logger=None,
            fifo_mode=True,
            delete_workdir=True,
            result_store=None,
            optional_dict=None,
            optional_dict2=None,
            parallelize=True,
            processes=1,
        )

        self.fassembler.run()

        results = self.fassembler.results

        self.assertAlmostEqual(results[0]['VMAF_feature_vif_score'], 0.44609306249999997, places=4)
        self.assertAlmostEqual(results[0]['VMAF_feature_motion_score'], 4.0498253541666669, places=4)
        self.assertAlmostEqual(results[0]['VMAF_feature_adm2_score'], 0.9345149030293786, places=4)
        self.assertAlmostEqual(results[0]['VMAF_feature_ansnr_score'], 23.509571520833333, places=4)

        self.assertAlmostEqual(results[1]['VMAF_feature_vif_score'], 1.0, places=4)
        self.assertAlmostEqual(results[1]['VMAF_feature_motion_score'], 4.0498253541666669, places=4)
        self.assertAlmostEqual(results[1]['VMAF_feature_adm2_score'], 1.0, places=4)
        self.assertAlmostEqual(results[1]['VMAF_feature_ansnr_score'], 31.271439270833337, places=4)

    def test_feature_assembler_selected_atom_feature(self):

        ref_path, dis_path, asset, asset_original = set_default_576_324_videos_for_testing()

        self.fassembler = FeatureAssembler(
            feature_dict={'VMAF_feature': ['vif', 'motion']},
            feature_option_dict=None,
            assets=[asset, asset_original],
            logger=None,
            fifo_mode=True,
            delete_workdir=True,
            result_store=None,
            optional_dict=None,
            optional_dict2=None,
            parallelize=True,
        )

        self.fassembler.run()

        results = self.fassembler.results

        self.assertAlmostEqual(results[0]['VMAF_feature_vif_score'], 0.44609306249999997, places=4)
        self.assertAlmostEqual(results[0]['VMAF_feature_motion_score'], 4.0498253541666669, places=4)

        self.assertAlmostEqual(results[1]['VMAF_feature_vif_score'], 1.0, places=4)
        self.assertAlmostEqual(results[1]['VMAF_feature_motion_score'], 4.0498253541666669, places=4)

        with self.assertRaises(KeyError):
            _ = results[0]['VMAF_feature_ansnr_scores']
        with self.assertRaises(KeyError):
            _ = results[0]['VMAF_feature_ansnr_score']
        with self.assertRaises(KeyError):
            _ = results[0]['VMAF_feature_adm_scores']
        with self.assertRaises(KeyError):
            _ = results[0]['VMAF_feature_adm_score']


class FeatureAssemblerUnitTest(unittest.TestCase):

    def test_feature_assembler_get_fextractor_instance(self):
        ref_path, dis_path, asset, asset_original = set_default_576_324_videos_for_testing()

        fassembler = FeatureAssembler(
            feature_dict={'VMAF_feature': ['vif_scale0', 'vif_scale1', 'vif_scale2', 'vif_scale3', 'adm2', 'motion2']},
            feature_option_dict={'VMAF_feature': {'adm_ref_display_height': 540}},
            assets=[asset],
            logger=None,
            fifo_mode=False,
            delete_workdir=True,
            result_store=None,
            optional_dict={'model_filepath': 'model/vmaf_float_v0.6.1_rdh540.json'},
            optional_dict2=None,
            parallelize=False,
            save_workfiles=False,
        )

        fex = fassembler._get_fextractor_instance('VMAF_feature')
        self.assertEqual(fex.optional_dict, {'adm_ref_display_height': 540})
        self.assertEqual(fex.optional_dict2, None)


if __name__ == '__main__':
    unittest.main(verbosity=2)



import copy
import hashlib

from slugify import slugify

from vmaf.tools.decorator import deprecated, override

__copyright__ = "Copyright 2016-2020, Netflix, Inc."
__license__ = "BSD+Patent"

import os

from vmaf.core.mixin import WorkdirEnabled
from vmaf.tools.misc import get_file_name_without_extension, \
    get_file_name_with_extension, get_unique_str_from_recursive_dict, \
    map_yuv_type_to_bitdepth
from vmaf.config import VmafConfig
from vmaf.core.proc_func import proc_func_dict


class Asset(WorkdirEnabled):
    """
    An Asset is the most basic unit with sufficient information to perform an
    execution task. It includes basic information about a distorted video and
    its undistorted reference video, as well as the frame range on which to
    extract features/calculate quality results (*dis_start_end_frame* and
    *ref_start_end_frame*), and at what resolution to perform such feature
    extraction (each video frame is upscaled to the resolution specified by
    *quality_width_height* before processing).

    Asset extends WorkdirEnabled mixin, which comes with a thread-safe working
    directory to facilitate parallel execution.

    The ref_path/dis_path points to the reference/distorted video files. For now,
    it supports YUV video files (yuvxxx), or encoded video files (notyuv) that
    can be decoded by ffmpeg.
    """

    SUPPORTED_YUV_TYPES = ['yuv420p', 'yuv422p', 'yuv444p',
                           'yuv420p10le', 'yuv422p10le', 'yuv444p10le',
                           'yuv420p12le', 'yuv422p12le', 'yuv444p12le',
                           'yuv420p16le', 'yuv422p16le', 'yuv444p16le',
                           'notyuv']
    DEFAULT_YUV_TYPE = 'yuv420p'

    SUPPORTED_RESAMPLING_TYPES = ['bilinear', 'bicubic', 'lanczos']
    DEFAULT_RESAMPLING_TYPE = 'bicubic'

    ORDERED_FILTER_LIST = ['crop', 'pad', 'gblur', 'eq', 'lutyuv', 'yadif']

    # ==== constructor ====

    def __init__(self, dataset, content_id, asset_id,
                 ref_path, dis_path,
                 asset_dict,
                 workdir_root=VmafConfig.workdir_path()):
        """
        :param dataset:
        :param content_id: ID of content the asset correspond to within dataset
        :param asset_id: ID of asset
        :param ref_path: path to reference video
        :param dis_path: path to distorted video
        :param asset_dict: dictionary with additional asset properties
        :param workdir_root:
        :return:
        """
        WorkdirEnabled.__init__(self, workdir_root)
        self.dataset = dataset
        self.content_id = content_id
        self.asset_id = asset_id
        self.ref_path = ref_path
        self.dis_path = dis_path
        self.asset_dict = asset_dict

        self._assert()

    def _assert(self):
        # validate yuv types
        assert self.ref_yuv_type in self.SUPPORTED_YUV_TYPES
        assert self.dis_yuv_type in self.SUPPORTED_YUV_TYPES
        assert self.workfile_yuv_type in self.SUPPORTED_YUV_TYPES
        # if YUV is notyuv, then ref/dis width and height should not be given,
        # since it must be encoded video and the information should be already
        # in included in the header of the video files
        if self.ref_yuv_type == 'notyuv':
            assert self.ref_width_height is None, 'For ref_yuv_type nonyuv, ref_width_height must NOT be specified.'
        if self.dis_yuv_type == 'notyuv':
            assert self.dis_width_height is None, 'For dis_yuv_type nonyuv, dis_width_height must NOT be specified.'
        # validate asset_dict
        self._assert_asset_dict()

    def _assert_asset_dict(self):
        # perform necessary assertions on asset properties of asset dict
        if 'fps' in self.asset_dict:
            assert self.asset_dict['fps'] > 0.0, 'Frame rate has to be positive.'
        if 'rebuf_indices' in self.asset_dict:
            assert isinstance(self.asset_dict['rebuf_indices'], list), 'Rebuffering indices need to be in a list.'
            # check for negative rebuffering indices
            assert len(list(filter(lambda x: x < 0, self.asset_dict['rebuf_indices']))) == 0, 'All rebuffering indices have to be >= 0.'

    def copy(self, **kwargs):
        new_asset_dict = copy.deepcopy(self.asset_dict)

        # reset the following arguments:
        if 'use_path_as_workpath' in new_asset_dict:
            del new_asset_dict['use_path_as_workpath']

        if 'use_workpath_as_procpath' in new_asset_dict:
            del new_asset_dict['use_workpath_as_procpath']

        dataset = kwargs['dataset'] if 'dataset' in kwargs else self.dataset
        content_id = kwargs['content_id'] if 'content_id' in kwargs else self.content_id
        asset_id = kwargs['asset_id'] if 'asset_id' in kwargs else self.asset_id
        ref_path = kwargs['ref_path'] if 'ref_path' in kwargs else self.ref_path
        dis_path = kwargs['dis_path'] if 'dis_path' in kwargs else self.dis_path
        workdir_root = kwargs['workdir_root'] if 'workdir_root' in kwargs else self.workdir_root

        # additional or override elements in asset_dict
        if 'asset_dict' in kwargs:
            for key in kwargs['asset_dict']:
                new_asset_dict[key] = kwargs['asset_dict'][key]

        new_asset = self.__class__(dataset, content_id, asset_id,
                                   ref_path, dis_path, new_asset_dict,
                                   workdir_root)
        return new_asset

    @classmethod
    def from_repr(cls, rp):
        """
        Reconstruct Asset from repr string.
        :return:
        """
        import ast
        d = ast.literal_eval(rp)
        assert 'dataset' in d
        assert 'content_id' in d
        assert 'asset_id' in d
        assert 'ref_path' in d
        assert 'dis_path' in d
        assert 'asset_dict' in d

        return cls(dataset=d['dataset'],
                   content_id=d['content_id'],
                   asset_id=d['asset_id'],
                   ref_path=d['ref_path'],
                   dis_path=d['dis_path'],
                   asset_dict=d['asset_dict']
                   )

    # ==== groundtruth ====
    @property
    def groundtruth(self):
        """
        Ground truth score, e.g. MOS, DMOS
        :return:
        """
        if 'groundtruth' in self.asset_dict:
            return self.asset_dict['groundtruth']
        else:
            return None

    @property
    def groundtruth_std(self):
        if 'groundtruth_std' in self.asset_dict:
            return self.asset_dict['groundtruth_std']
        else:
            return None

    @property
    def raw_groundtruth(self):
        """
        Raw ground truth scores, e.g. opinion score (OS)
        :return:
        """
        if 'raw_groundtruth' in self.asset_dict:
            return self.asset_dict['raw_groundtruth']
        else:
            return None

    # ==== width and height ====

    @property
    def ref_width_height(self):
        """
        Width and height of reference video.
        :return: width and height of reference video. If None, it signals that
        width and height should be figured out in other means (e.g. FFMPEG).
        """
        if 'ref_width' in self.asset_dict and 'ref_height' in self.asset_dict:
            return self.asset_dict['ref_width'], self.asset_dict['ref_height']
        elif 'width' in self.asset_dict and 'height' in self.asset_dict:
            return self.asset_dict['width'], self.asset_dict['height']
        else:
            return None

    @property
    def dis_width_height(self):
        """
        Width and height of distorted video.
        :return: width and height of distorted video. If None, it signals that
        width and height should be figured out in other means (e.g. FFMPEG)
        """
        if 'dis_width' in self.asset_dict and 'dis_height' in self.asset_dict:
            return self.asset_dict['dis_width'], self.asset_dict['dis_height']
        elif 'width' in self.asset_dict and 'height' in self.asset_dict:
            return self.asset_dict['width'], self.asset_dict['height']
        else:
            return None

    @property
    def dis_encode_width_height(self):
        """
        Width and height of the encoded video before any scaling was applied.
        :return: width and height at encode resolution.
        Defaults to dis_width_height.
        """

        if 'dis_enc_width' in self.asset_dict and 'dis_enc_height' in self.asset_dict:
            return self.asset_dict['dis_enc_width'], self.asset_dict['dis_enc_height']
        else:
            return self.dis_width_height

    @property
    def dis_encode_bitdepth(self):
        """
        Bitdepth of the encoded video before any conversions were applied.
        :return: bitdepth of the encode.
        Defaults to bitdepth of dis_yuv_type (e.g. 8 for yuv420p).
        """
        if 'dis_enc_bitdepth' in self.asset_dict:
            assert self.asset_dict['dis_enc_bitdepth'] in [8, 10, 12, 16], \
                "Supported encoding bitdepths are 8, 10, 12, and 16."
            return self.asset_dict['dis_enc_bitdepth']
        else:
            return map_yuv_type_to_bitdepth(self.dis_yuv_type)

    def clear_up_width_height(self):
        if 'width' in self.asset_dict:
            del self.asset_dict['width']
        if 'height' in self.asset_dict:
            del self.asset_dict['height']
        if 'ref_width' in self.asset_dict:
            del self.asset_dict['ref_width']
        if 'ref_height' in self.asset_dict:
            del self.asset_dict['ref_height']
        if 'dis_width' in self.asset_dict:
            del self.asset_dict['dis_width']
        if 'dis_height' in self.asset_dict:
            del self.asset_dict['dis_height']

    @property
    def quality_width_height(self):
        """
        Width and height to scale distorted video to before quality calculation.
        :return: width and height at which the quality is measured at. either
        'quality_width' and 'quality_height' have to present in asset_dict;
        or ref and dis's width and height must be equal, which will be used
        as the default quality width and height; or either of ref/dis is type
        'notyuv', in which case the other's width/height (could also be None)
        """

        if 'quality_width' in self.asset_dict and 'quality_height' in self.asset_dict:
            return self.asset_dict['quality_width'], self.asset_dict['quality_height']
        elif self.ref_yuv_type == 'notyuv':
            return self.dis_width_height
        elif self.dis_yuv_type == 'notyuv':
            return self.ref_width_height
        else:
            assert self.ref_width_height == self.dis_width_height
            return self.ref_width_height

    # ==== start and end frame ====

    @property
    def ref_start_end_frame(self):
        """
        Start and end frame of reference video for quality calculation.
        :return: reference video's start frame and end frame for processing
        (inclusive). If None, it signals that the entire video should be
        processed.
        """
        if 'ref_start_frame' in self.asset_dict and 'ref_end_frame' in self.asset_dict:
            return self.asset_dict['ref_start_frame'], self.asset_dict['ref_end_frame']

        elif 'start_frame' in self.asset_dict and 'end_frame' in self.asset_dict:
            return self.asset_dict['start_frame'], self.asset_dict['end_frame']

        elif 'start_sec' in self.asset_dict and 'end_sec' in self.asset_dict and 'fps' in self.asset_dict:
            start_frame = int(round(self.asset_dict['start_sec'] * self.asset_dict['fps']))
            end_frame = int(round(self.asset_dict['end_sec'] * self.asset_dict['fps'])) - 1
            return start_frame, end_frame

        elif 'duration_sec' in self.asset_dict and 'fps' in self.asset_dict:
            start_frame = 0
            end_frame = int(round(self.asset_dict['duration_sec'] * self.asset_dict['fps'])) - 1
            return start_frame, end_frame

        else:
            return None

    @property
    def dis_start_end_frame(self):
        """
        Start and end frame of distorted video for quality calculation.
        :return: distorted video's start frame and end frame for processing
        (inclusive). If None, it signals that the entire video should be
        processed.
        """
        if 'dis_start_frame' in self.asset_dict and 'dis_end_frame' in self.asset_dict:
            return self.asset_dict['dis_start_frame'], self.asset_dict['dis_end_frame']

        elif 'start_frame' in self.asset_dict and 'end_frame' in self.asset_dict:
            return self.asset_dict['start_frame'], self.asset_dict['end_frame']

        elif 'start_sec' in self.asset_dict and 'end_sec' in self.asset_dict and 'fps' in self.asset_dict:
            start_frame = int(round(self.asset_dict['start_sec'] * self.asset_dict['fps']))
            end_frame = int(round(self.asset_dict['end_sec'] * self.asset_dict['fps'])) - 1
            return start_frame, end_frame

        elif 'duration_sec' in self.asset_dict and 'fps' in self.asset_dict:
            start_frame = 0
            end_frame = int(round(self.asset_dict['duration_sec'] * self.asset_dict['fps'])) - 1
            return start_frame, end_frame

        else:
            return None

    def clear_up_start_end_frame(self):
        if 'start_frame' in self.asset_dict:
            del self.asset_dict['start_frame']
        if 'end_frame' in self.asset_dict:
            del self.asset_dict['end_frame']
        if 'ref_start_frame' in self.asset_dict:
            del self.asset_dict['ref_start_frame']
        if 'dis_start_frame' in self.asset_dict:
            del self.asset_dict['dis_start_frame']
        if 'start_sec' in self.asset_dict:
            del self.asset_dict['start_sec']
        if 'end_sec' in self.asset_dict:
            del self.asset_dict['end_sec']
        if 'duration_sec' in self.asset_dict:
            del self.asset_dict['duration_sec']

    # ==== duration and start time====

    @property
    def ref_duration_sec(self):
        """
        Reference video's duration in second used in quality calculation.
        :return:
        """
        if 'duration_sec' in self.asset_dict:
            return self.asset_dict['duration_sec']
        elif 'start_sec' in self.asset_dict \
                and 'end_sec' in self.asset_dict:
            return self.asset_dict['end_sec'] - self.asset_dict['start_sec']
        else:
            ref_start_end_frame = self.ref_start_end_frame
            if ref_start_end_frame and 'fps' in self.asset_dict:
                s, e = ref_start_end_frame
                return (e - s + 1) / float(self.asset_dict['fps'])
            else:
                return None

    @property
    def dis_duration_sec(self):
        """
        Distorted video's duration in second used in quality calculation.
        :return:
        """
        if 'duration_sec' in self.asset_dict:
            return self.asset_dict['duration_sec']
        elif 'start_sec' in self.asset_dict \
                and 'end_sec' in self.asset_dict:
            return self.asset_dict['end_sec'] - self.asset_dict['start_sec']
        else:
            dis_start_end_frame = self.dis_start_end_frame
            if dis_start_end_frame \
                    and 'fps' in self.asset_dict:
                start, end = dis_start_end_frame
                return (end - start + 1) / float(self.asset_dict['fps'])
            else:
                return None

    @property
    def ref_start_sec(self):
        if self.ref_start_end_frame is None or self.fps is None:
            return None
        else:
            ref_start_frame, ref_end_frame = self.ref_start_end_frame
            fps = self.fps
            return float(ref_start_frame) / fps

    @property
    def dis_start_sec(self):
        if self.dis_start_end_frame is None or self.fps is None:
            return None
        else:
            dis_start_frame, dis_end_frame = self.dis_start_end_frame
            fps = self.fps
            return float(dis_start_frame) / fps

    @property
    def fps(self):
        if 'fps' in self.asset_dict:
            assert self.asset_dict['fps'] > 0.0, 'Frame rate has to be positive.'
            return self.asset_dict['fps']
        else:
            return None

    @property
    def rebuf_indices(self):
        if 'rebuf_indices' in self.asset_dict:
            assert isinstance(self.asset_dict['rebuf_indices'], list), 'Rebuffering indices need to be in a list.'
            # check for negative rebuffering indices
            assert len(list(filter(lambda x: x < 0, self.asset_dict['rebuf_indices']))) == 0, 'All rebuffering indices have to be >= 0.'
            return self.asset_dict['rebuf_indices']
        else:
            return None

    # ==== str ====

    @property
    def ref_str(self):
        """
        String representation for reference video.
        :return:
        """
        s = ""

        path = get_file_name_without_extension(self.ref_path)
        s += "{path}".format(path=path)

        if self.ref_width_height:
            w, h = self.ref_width_height
            s += "_{w}x{h}".format(w=w, h=h)

        if self.ref_yuv_type != self.DEFAULT_YUV_TYPE:
            s += "_{}".format(self.ref_yuv_type)

        # if resolutions are consistent, no resampling is taking place, so
        # specificying resampling type should be ignored
        if self.ref_resampling_type != self.DEFAULT_RESAMPLING_TYPE and \
                not self.ref_width_height == self.quality_width_height:
            if s != "":
                s += "_"
            s += "{}".format(self.ref_resampling_type)

        if self.ref_start_end_frame:
            start, end = self.ref_start_end_frame
            s += "_{start}to{end}".format(start=start, end=end)

        for key in self.ORDERED_FILTER_LIST:
            if self.get_filter_cmd(key, 'ref') is not None:
                if s != "":
                    s += "_"
                s += "{}{}".format(key, self.get_filter_cmd(key, 'ref'))

        if self.ref_proc_callback_str:
            s += f'_{self.ref_proc_callback_str}'

        return slugify(s, separator='_')

    @property
    def dis_str(self):
        """
        String representation for distorted video.
        :return:
        """
        s = ""

        path = get_file_name_without_extension(self.dis_path)
        s += "{path}".format(path=path)

        if self.dis_width_height:
            w, h = self.dis_width_height
            s += "_{w}x{h}".format(w=w, h=h)

        if self.dis_encode_width_height != self.dis_width_height:
            w, h = self.dis_encode_width_height
            s += "_e_{w}x{h}".format(w=w, h=h)

        if self.dis_encode_bitdepth is not None and \
                map_yuv_type_to_bitdepth(self.dis_yuv_type) != self.dis_encode_bitdepth:
            # only add dis_encode_bitdepth to the string if it is not None, and it is different from the bitdepth
            # of dis_yuv_type
            ebd = self.dis_encode_bitdepth
            s += "_ebd_{ebd}".format(ebd=ebd)

        if self.dis_yuv_type != self.DEFAULT_YUV_TYPE:
            s += "_{}".format(self.dis_yuv_type)

        # if resolutions are consistent, no resampling is taking place, so
        # specificying resampling type should be ignored
        if self.dis_resampling_type != self.DEFAULT_RESAMPLING_TYPE and \
                not self.dis_width_height == self.quality_width_height:
            if s != "":
                s += "_"
            s += "{}".format(self.dis_resampling_type)

        if self.dis_start_end_frame:
            start, end = self.dis_start_end_frame
            s += "_{start}to{end}".format(start=start, end=end)

        for key in self.ORDERED_FILTER_LIST:
            if self.get_filter_cmd(key, 'dis') is not None:
                if s != "":
                    s += "_"
                s += "{}{}".format(key, self.get_filter_cmd(key, 'dis'))

        if self.dis_proc_callback_str:
            s += f'_{self.dis_proc_callback_str}'

        return slugify(s, separator='_')

    @property
    def quality_str(self):
        """
        String representation for quality-related information
        :return:
        """
        s = ""

        if self.quality_width_height:
            w, h = self.quality_width_height
            if s != "":
                s += "_"
            s += "{w}x{h}".format(w=w, h=h)

        if self.workfile_yuv_type != self.DEFAULT_YUV_TYPE:
            if s != "":
                s += "_"
            s += "wf_" + self.workfile_yuv_type

        return s

    def to_string(self):
        """
        The compact string representation of asset, used by __str__.
        :return:
        """
        s = "{dataset}_{content_id}_{asset_id}_{ref_str}_vs_{dis_str}".\
            format(dataset=self.dataset,
                   content_id=self.content_id,
                   asset_id=self.asset_id,
                   ref_str=self.ref_str,
                   dis_str=self.dis_str)
        quality_str = self.quality_str
        if quality_str:
            s += "_q_{quality_str}".format(quality_str=quality_str)

        if len(s) > 196:  # upper limit of filename is 256 but leave some space for prefix/suffix
            s = hashlib.sha1(s.encode("utf-8")).hexdigest()

        return s

    def to_normalized_dict(self):
        """
        Similar to self.__dict__ except for excluding workdir (which is random)
        and dir part of ref_path/dis_path.
        :return:
        """
        d = {}
        for key in self.__dict__:
            if key == 'workdir':
                d[key] = ""
            elif key == 'ref_path' or key == 'dis_path':
                d[key] = get_file_name_with_extension(self.__dict__[key])
            else:
                d[key] = self.__dict__[key]
        return d

    def __str__(self):
        """
        Use str(asset) for compact but unique description of asset, for example
        use in file names
        :return:
        """
        return self.to_string()

    def __repr__(self):
        """
        Use repr(asset) for serialization of asset (to be recovered later on)
        :return:
        """
        return self.to_normalized_repr()

    def to_full_repr(self):
        return get_unique_str_from_recursive_dict(self.__dict__)

    def to_normalized_repr(self):
        return get_unique_str_from_recursive_dict(self.to_normalized_dict())

    def __hash__(self):
        return hash(self.to_normalized_repr())

    def __eq__(self, other):
        return self.to_normalized_repr() == other.to_normalized_repr()

    def __ne__(self, other):
        return not self.__eq__(other)

    # ==== workfile ====

    @property
    def ref_workfile_path(self):
        if self.use_path_as_workpath:
            return self.ref_path
        else:
            return os.path.join(self.workdir, f"ref_{str(self)}")

    @property
    def dis_workfile_path(self):
        if self.use_path_as_workpath:
            return self.dis_path
        else:
            return os.path.join(self.workdir, f"dis_{str(self)}")

    # ==== procfile ====

    @property
    def ref_procfile_path(self):
        if self.use_workpath_as_procpath:
            return self.ref_workfile_path
        else:
            return os.path.join(self.workdir, f"refp_{str(self)}")

    @property
    def dis_procfile_path(self):
        if self.use_workpath_as_procpath:
            return self.dis_workfile_path
        else:
            return os.path.join(self.workdir, f"disp_{str(self)}")

    # ==== bitrate ====

    @property
    def ref_bitrate_kbps_for_entire_file(self):
        """
        :return: the bitrate in Kbps for the entire reference video file. Must
        make sure ref_duration_sec covers the entire file.
        """
        try:
            return os.path.getsize(self.ref_path) / self.ref_duration_sec * 8.0 / 1000.0
        except:
            return None

    @property
    def dis_bitrate_kbps_for_entire_file(self):
        """
        :return: the bitrate in Kbps for the entire reference video file. Must
        make sure ref_duration_sec covers the entire file.
        """
        try:
            return os.path.getsize(self.dis_path) / self.dis_duration_sec * 8.0 / 1000.0
        except:
            return None

    # ==== yuv format ====

    @property
    def ref_yuv_type(self):
        if 'ref_yuv_type' in self.asset_dict:
            if self.asset_dict['ref_yuv_type'] in self.SUPPORTED_YUV_TYPES:
                return self.asset_dict['ref_yuv_type']
            else:
                assert False, "Unsupported YUV type: {}".format(
                    self.asset_dict['ref_yuv_type'])
        elif 'yuv_type' in self.asset_dict:
            if self.asset_dict['yuv_type'] in self.SUPPORTED_YUV_TYPES:
                return self.asset_dict['yuv_type']
            else:
                assert False, "Unsupported YUV type: {}".format(
                    self.asset_dict['yuv_type'])
        else:
            return self.DEFAULT_YUV_TYPE

    @property
    def dis_yuv_type(self):
        if 'dis_yuv_type' in self.asset_dict:
            if self.asset_dict['dis_yuv_type'] in self.SUPPORTED_YUV_TYPES:
                return self.asset_dict['dis_yuv_type']
            else:
                assert False, "Unsupported YUV type: {}".format(
                    self.asset_dict['dis_yuv_type'])
        elif 'yuv_type' in self.asset_dict:
            if self.asset_dict['yuv_type'] in self.SUPPORTED_YUV_TYPES:
                return self.asset_dict['yuv_type']
            else:
                assert False, "Unsupported YUV type: {}".format(
                    self.asset_dict['yuv_type'])
        else:
            return self.DEFAULT_YUV_TYPE

    @property
    @deprecated
    def yuv_type(self):
        """ For backward-compatibility """
        assert self.ref_yuv_type == self.dis_yuv_type
        return self.dis_yuv_type

    @property
    def workfile_yuv_type(self):
        """
        for notyuv assets, we want to allow the decoded yuv format to be set by the user
        this is highly relevant to image decoding, where we would like to select yuv444p
        this property tries to read workfile_yuv_type from asset_dict, if it is there it is set
        else it defaults to default_yuv_type
        """
        supported_yuv_types = list(set(Asset.SUPPORTED_YUV_TYPES) - {'notyuv'})
        if 'workfile_yuv_type' in self.asset_dict:
            workfile_yuv_type = self.asset_dict['workfile_yuv_type']
            assert workfile_yuv_type in supported_yuv_types, "Workfile YUV format {} is not valid, pick: {}".format(
                workfile_yuv_type, str(supported_yuv_types))
            return workfile_yuv_type
        else:
            return self.DEFAULT_YUV_TYPE

    def clear_up_yuv_type(self):
        if 'yuv_type' in self.asset_dict:
            del self.asset_dict['yuv_type']
        if 'ref_yuv_type' in self.asset_dict:
            del self.asset_dict['ref_yuv_type']
        if 'dis_yuv_type' in self.asset_dict:
            del self.asset_dict['dis_yuv_type']

    # ==== resampling type ====

    @property
    def ref_resampling_type(self):
        if 'ref_resampling_type' in self.asset_dict:
            if self.asset_dict['ref_resampling_type'] in self.SUPPORTED_RESAMPLING_TYPES:
                return self.asset_dict['ref_resampling_type']
            else:
                assert False, "Unsupported resampling type: {}".format(
                    self.asset_dict['ref_resampling_type'])
        elif 'resampling_type' in self.asset_dict:
            if self.asset_dict['resampling_type'] in self.SUPPORTED_RESAMPLING_TYPES:
                return self.asset_dict['resampling_type']
            else:
                assert False, "Unsupported resampling type: {}".format(
                    self.asset_dict['resampling_type'])
        else:
            return self.DEFAULT_RESAMPLING_TYPE

    @property
    def dis_resampling_type(self):
        if 'dis_resampling_type' in self.asset_dict:
            if self.asset_dict['dis_resampling_type'] in self.SUPPORTED_RESAMPLING_TYPES:
                return self.asset_dict['dis_resampling_type']
            else:
                assert False, "Unsupported resampling type: {}".format(
                    self.asset_dict['dis_resampling_type'])
        elif 'resampling_type' in self.asset_dict:
            if self.asset_dict['resampling_type'] in self.SUPPORTED_RESAMPLING_TYPES:
                return self.asset_dict['resampling_type']
            else:
                assert False, "Unsupported resampling type: {}".format(
                    self.asset_dict['resampling_type'])
        else:
            return self.DEFAULT_RESAMPLING_TYPE

    @property
    @deprecated
    def resampling_type(self):
        """ For backward-compatibility """
        assert self.ref_resampling_type == self.dis_resampling_type
        return self.dis_resampling_type

    @property
    def use_path_as_workpath(self):
        """
        If True, use ref_path as ref_workfile_path, and dis_path as
        dis_workfile_path.
        """
        if 'use_path_as_workpath' in self.asset_dict:
            if self.asset_dict['use_path_as_workpath'] == 1:
                return True
            elif self.asset_dict['use_path_as_workpath'] == 0:
                return False
            else:
                assert False
        else:
            return False

    @use_path_as_workpath.setter
    def use_path_as_workpath(self, bool_value):
        # cannot just assign True/False for ResultStore reason:
        # df = pd.DataFrame.from_dict(ast.literal_eval(result_file.read()))
        # cannot read true/false
        if bool_value is True:
            self.asset_dict['use_path_as_workpath'] = 1
        else:
            self.asset_dict['use_path_as_workpath'] = 0

    @property
    def use_workpath_as_procpath(self):
        """
        If True, use ref_workfile_path as ref_procfile_path, and dis_workfile_path
        as dis_procfile_path.
        """
        if 'use_workpath_as_procpath' in self.asset_dict:
            if self.asset_dict['use_workpath_as_procpath'] == 1:
                return True
            elif self.asset_dict['use_workpath_as_procpath'] == 0:
                return False
            else:
                assert False
        else:
            return False

    @use_workpath_as_procpath.setter
    def use_workpath_as_procpath(self, bool_value):
        # cannot just assign True/False for ResultStore reason:
        # df = pd.DataFrame.from_dict(ast.literal_eval(result_file.read()))
        # cannot read true/false
        if bool_value is True:
            self.asset_dict['use_workpath_as_procpath'] = 1
        else:
            self.asset_dict['use_workpath_as_procpath'] = 0

    @property
    def crop_cmd(self):
        return self.get_filter_cmd('crop', None)

    @property
    def ref_crop_cmd(self):
        return self.get_filter_cmd('crop', 'ref')

    @property
    def dis_crop_cmd(self):
        return self.get_filter_cmd('crop', 'dis')

    @property
    def pad_cmd(self):
        return self.get_filter_cmd('pad', None)

    @property
    def ref_pad_cmd(self):
        return self.get_filter_cmd('pad', 'ref')

    @property
    def dis_pad_cmd(self):
        return self.get_filter_cmd('pad', 'dis')

    def get_filter_cmd(self, key, target=None):
        assert key in self.ORDERED_FILTER_LIST, 'key {key} is not in SUPPORTED_FILTER_TYPES'.format(key=key)
        assert target == 'ref' or target == 'dis' or target is None, \
            'target is {}, which is not supported'.format(target)
        if target is None:
            cmd = key + '_cmd'
            if cmd in self.asset_dict:
                return self.asset_dict[cmd]
            else:
                return None
        if target == 'ref' or target == 'dis':
            cmd = target + '_' + key + '_cmd'
            cmd2 = key + '_cmd'
            if cmd in self.asset_dict:
                return self.asset_dict[cmd]
            elif cmd2 in self.asset_dict:
                return self.asset_dict[cmd2]
            else:
                return None
        else:
            assert False

    @property
    def ref_proc_callback_str(self):
        if 'ref_proc_callback' in self.asset_dict:
            if self.asset_dict['ref_proc_callback'] in proc_func_dict:
                return self.asset_dict['ref_proc_callback']
            else:
                assert False, "Unsupported ref_proc_callback: {}".format(
                    self.asset_dict['ref_proc_callback'])
        elif 'proc_callback' in self.asset_dict:
            if self.asset_dict['proc_callback'] in proc_func_dict:
                return self.asset_dict['proc_callback']
            else:
                assert False, "Unsupported proc_callback: {}".format(
                    self.asset_dict['proc_callback'])
        else:
            return None

    @property
    def ref_proc_callback(self):
        if self.ref_proc_callback_str is None:
            return None
        else:
            return proc_func_dict[self.ref_proc_callback_str]

    @property
    def dis_proc_callback_str(self):
        if 'dis_proc_callback' in self.asset_dict:
            if self.asset_dict['dis_proc_callback'] in proc_func_dict:
                return self.asset_dict['dis_proc_callback']
            else:
                assert False, "Unsupported dis_proc_callback: {}".format(
                    self.asset_dict['dis_proc_callback'])
        elif 'proc_callback' in self.asset_dict:
            if self.asset_dict['proc_callback'] in proc_func_dict:
                return self.asset_dict['proc_callback']
            else:
                assert False, "Unsupported proc_callback: {}".format(
                    self.asset_dict['proc_callback'])
        else:
            return None

    @property
    def dis_proc_callback(self):
        if self.dis_proc_callback_str is None:
            return None
        else:
            return proc_func_dict[self.dis_proc_callback_str]


class NorefAsset(Asset):
    """
    NorefAsset is similar to Asset except that it does not have a reference
    video path ref_path.
    """

    # ==== constructor ====

    def __init__(self, dataset, content_id, asset_id,
                 dis_path,
                 asset_dict,
                 workdir_root=VmafConfig.workdir_path()):
        """
        :param dataset:
        :param content_id: ID of content the asset correspond to within dataset
        :param asset_id: ID of asset
        :param dis_path: path to distorted video
        :param asset_dict: dictionary with additional asset properties
        :param workdir_root:
        :return:
        """
        super(NorefAsset, self).__init__(
            dataset,
            content_id,
            asset_id,
            dis_path,  # repeat dis_path for both ref_path and dis_path
            dis_path,
            asset_dict,
            workdir_root
        )

    @override(Asset)
    def copy(self, **kwargs):
        new_asset_dict = copy.deepcopy(self.asset_dict)

        # reset the following arguments:
        if 'use_path_as_workpath' in new_asset_dict:
            del new_asset_dict['use_path_as_workpath']

        if 'use_workpath_as_procpath' in new_asset_dict:
            del new_asset_dict['use_workpath_as_procpath']

        dataset = kwargs['dataset'] if 'dataset' in kwargs else self.dataset
        content_id = kwargs['content_id'] if 'content_id' in kwargs else self.content_id
        asset_id = kwargs['asset_id'] if 'asset_id' in kwargs else self.asset_id
        dis_path = kwargs['dis_path'] if 'dis_path' in kwargs else self.dis_path
        workdir_root = kwargs['workdir_root'] if 'workdir_root' in kwargs else self.workdir_root

        # additional or override elements in asset_dict
        if 'asset_dict' in kwargs:
            for key in kwargs['asset_dict']:
                new_asset_dict[key] = kwargs['asset_dict'][key]

        new_asset = self.__class__(dataset, content_id, asset_id,
                                   dis_path, new_asset_dict,
                                   workdir_root)
        return new_asset

    def copy_as_asset(self, **kwargs):
        """ similar to Noref.copy, except that the returned object is of
        (super)class Asset. """
        new_asset = self.copy()
        new_asset.__class__ = Asset
        return new_asset.copy(**kwargs)

    @property
    @override(Asset)
    def ref_width_height(self):
        return self.dis_width_height

    @property
    @override(Asset)
    def ref_start_end_frame(self):
        return self.dis_start_end_frame

    @property
    @override(Asset)
    def ref_duration_sec(self):
        return self.dis_duration_sec

    @property
    @override
    def ref_start_sec(self):
        return self.dis_start_sec

    @property
    @override(Asset)
    def ref_yuv_type(self):
        return self.dis_yuv_type

    @property
    @override(Asset)
    def ref_resampling_type(self):
        return self.dis_resampling_type

    @property
    @override(Asset)
    def ref_crop_cmd(self):
        return self.dis_crop_cmd

    @property
    @override(Asset)
    def ref_pad_cmd(self):
        return self.dis_pad_cmd

    @property
    @override(Asset)
    def ref_proc_callback_str(self):
        return self.dis_proc_callback_str

    @property
    @override(Asset)
    def ref_proc_callback(self):
        return self.dis_proc_callback

    def to_string(self):
        """
        The compact string representation of asset, used by __str__.
        :return:
        """
        s = "{dataset}_{content_id}_{asset_id}_{dis_str}". \
            format(dataset=self.dataset,
                   content_id=self.content_id,
                   asset_id=self.asset_id,
                   dis_str=self.dis_str)
        quality_str = self.quality_str
        if quality_str:
            s += "_q_{quality_str}".format(quality_str=quality_str)

        if len(s) > 196:  # upper limit of filename is 256 but leave some space for prefix/suffix
            s = hashlib.sha1(s.encode("utf-8")).hexdigest()

        return s



from __future__ import absolute_import

import unittest
from functools import partial

import numpy as np

from vmaf.core.noref_feature_extractor import MomentNorefFeatureExtractor, \
    NiqeNorefFeatureExtractor, BrisqueNorefFeatureExtractor, SiTiNorefFeatureExtractor

from test.testutil import set_default_576_324_videos_for_testing, set_default_576_324_noref_videos_for_testing

__copyright__ = "Copyright 2016-2020, Netflix, Inc."
__license__ = "BSD+Patent"

from vmaf.core.result_store import FileSystemResultStore

from vmaf.tools.misc import MyTestCase


class NorefFeatureExtractorTest(unittest.TestCase):

    def tearDown(self):
        if hasattr(self, 'fextractor'):
            self.fextractor.remove_results()
            pass

    def test_noref_moment_fextractor(self):

        ref_path, dis_path, asset, asset_original = set_default_576_324_videos_for_testing()

        self.fextractor = MomentNorefFeatureExtractor(
            [asset, asset_original],
            None, fifo_mode=True,
            result_store=None
        )
        self.fextractor.run()

        results = self.fextractor.results

        self.assertAlmostEqual(results[0]['Moment_noref_feature_1st_score'], 61.332006624999984)
        self.assertAlmostEqual(results[0]['Moment_noref_feature_2nd_score'], 4798.659574041666)
        self.assertAlmostEqual(results[0]['Moment_noref_feature_var_score'], 1036.8371843488285)

        self.assertAlmostEqual(results[1]['Moment_noref_feature_1st_score'], 59.788567297525134)
        self.assertAlmostEqual(results[1]['Moment_noref_feature_2nd_score'], 4696.668388042271)
        self.assertAlmostEqual(results[1]['Moment_noref_feature_var_score'], 1121.519917231207)

    def test_noref_moment_fextractor_with_noref_asset(self):

        ref_path, dis_path, asset, asset_original = set_default_576_324_noref_videos_for_testing()

        self.fextractor = MomentNorefFeatureExtractor(
            [asset, asset_original],
            None, fifo_mode=True,
            result_store=None
        )
        self.fextractor.run(parallelize=True)

        results = self.fextractor.results

        self.assertAlmostEqual(results[0]['Moment_noref_feature_1st_score'], 61.332006624999984)
        self.assertAlmostEqual(results[0]['Moment_noref_feature_2nd_score'], 4798.659574041666)
        self.assertAlmostEqual(results[0]['Moment_noref_feature_var_score'], 1036.8371843488285)

        self.assertAlmostEqual(results[1]['Moment_noref_feature_1st_score'], 59.788567297525134)
        self.assertAlmostEqual(results[1]['Moment_noref_feature_2nd_score'], 4696.668388042271)
        self.assertAlmostEqual(results[1]['Moment_noref_feature_var_score'], 1121.519917231207)

    def test_run_noref_brisque_fextractor(self):

        ref_path, dis_path, asset, asset_original = set_default_576_324_videos_for_testing()

        self.fextractor = BrisqueNorefFeatureExtractor(
            [asset, asset_original],
            None, fifo_mode=True,
            result_store=None
        )

        self.fextractor.run(parallelize=True)

        results = self.fextractor.results

        self.assertAlmostEqual(results[0]['BRISQUE_noref_feature_alpha23_score'], 0.7640625000000005, places=4)
        self.assertAlmostEqual(results[0]['BRISQUE_noref_feature_alpha13_score'], 0.6322500000000002, places=4)
        self.assertAlmostEqual(results[0]['BRISQUE_noref_feature_N34_score'],    -0.007239876204980851, places=4)

        self.assertAlmostEqual(results[1]['BRISQUE_noref_feature_alpha23_score'], 0.8644583333333339, places=4)
        self.assertAlmostEqual(results[1]['BRISQUE_noref_feature_alpha13_score'], 0.82906250000000103, places=4)
        self.assertAlmostEqual(results[1]['BRISQUE_noref_feature_N34_score'],     -0.0092448158862212092, places=4)

    def test_run_noref_niqe_fextractor(self):

        ref_path, dis_path, asset, asset_original = set_default_576_324_videos_for_testing()

        self.fextractor = NiqeNorefFeatureExtractor(
            [asset, asset_original],
            None, fifo_mode=False,
            result_store=None
        )

        self.fextractor.run(parallelize=True)

        results = self.fextractor.results

        self.assertAlmostEqual(results[0]['NIQE_noref_feature_alpha23_score'], 0.8168807870370377, places=4)
        self.assertAlmostEqual(results[0]['NIQE_noref_feature_alpha13_score'], 0.6949641203703707, places=4)
        self.assertAlmostEqual(results[0]['NIQE_noref_feature_alpha_m1_score'], 2.0924143518518536, places=3)
        self.assertAlmostEqual(results[0]['NIQE_noref_feature_blbr1_score'], 0.72958634325785898, places=3)

        self.assertAlmostEqual(results[1]['NIQE_noref_feature_alpha23_score'], 0.89566087962963026, places=4)
        self.assertAlmostEqual(results[1]['NIQE_noref_feature_alpha13_score'], 0.85539583333333391, places=4)
        self.assertAlmostEqual(results[1]['NIQE_noref_feature_alpha_m1_score'], 2.7192025462962985, places=4)
        self.assertAlmostEqual(results[1]['NIQE_noref_feature_blbr1_score'], 0.98723051960738684, places=4)

    def test_run_noref_niqe_fextractor_train(self):

        ref_path, dis_path, asset, asset_original = set_default_576_324_videos_for_testing()

        self.fextractor = NiqeNorefFeatureExtractor(
            [asset, asset_original],
            None, fifo_mode=False,
            result_store=None,
            optional_dict={'mode': 'train'},
            optional_dict2=None,
        )

        self.fextractor.run(parallelize=True)

        results = self.fextractor.results

        self.assertAlmostEqual(results[0]['NIQE_noref_feature_alpha23_score'], 0.97259000000000073, places=4)
        self.assertAlmostEqual(results[0]['NIQE_noref_feature_alpha13_score'], 0.80907000000000051, places=4)
        self.assertAlmostEqual(results[0]['NIQE_noref_feature_alpha_m1_score'], 2.6135250000000019, places=4)
        self.assertAlmostEqual(results[0]['NIQE_noref_feature_blbr1_score'], 0.9150526409258144, places=4)

        self.assertAlmostEqual(results[1]['NIQE_noref_feature_alpha23_score'], 0.97447727272727347, places=4)
        self.assertAlmostEqual(results[1]['NIQE_noref_feature_alpha13_score'], 0.89120909090909162, places=4)
        self.assertAlmostEqual(results[1]['NIQE_noref_feature_alpha_m1_score'], 3.0300909090909118, places=4)
        self.assertAlmostEqual(results[1]['NIQE_noref_feature_blbr1_score'], 1.0508255408831713, places=4)

    def test_run_noref_niqe_fextractor_with_patch_size(self):

        ref_path, dis_path, asset, asset_original = set_default_576_324_videos_for_testing()

        self.fextractor = NiqeNorefFeatureExtractor(
            [asset, asset_original],
            None, fifo_mode=False,
            result_store=None,
            optional_dict={'patch_size': 48},
            optional_dict2=None,
        )

        self.fextractor.run(parallelize=True)

        results = self.fextractor.results

        self.assertAlmostEqual(results[0]['NIQE_noref_feature_alpha23_score'], 0.8430156250000006, places=4)
        self.assertAlmostEqual(results[0]['NIQE_noref_feature_alpha13_score'], 0.71714583333333359, places=4)
        self.assertAlmostEqual(results[0]['NIQE_noref_feature_alpha_m1_score'], 2.2195590277777795, places=3)
        self.assertAlmostEqual(results[0]['NIQE_noref_feature_blbr1_score'], 0.74061215376929412, places=3)

        self.assertAlmostEqual(results[1]['NIQE_noref_feature_alpha23_score'], 0.9144918981481488, places=4)
        self.assertAlmostEqual(results[1]['NIQE_noref_feature_alpha13_score'], 0.87132291666666728, places=4)
        self.assertAlmostEqual(results[1]['NIQE_noref_feature_alpha_m1_score'], 2.8193532986111136, places=4)
        self.assertAlmostEqual(results[1]['NIQE_noref_feature_blbr1_score'], 0.99354006450609134, places=4)

    def test_noref_siti_fextractor_with_noref_asset(self):

        ref_path, dis_path, asset, asset_original = set_default_576_324_noref_videos_for_testing()

        self.fextractor = SiTiNorefFeatureExtractor(
            [asset, asset_original],
            None, fifo_mode=True,
            result_store=None
        )
        self.fextractor.run(parallelize=True)

        results = self.fextractor.results

        self.assertAlmostEqual(results[0]['SITI_noref_feature_si_score'], 79.27042052471928)
        self.assertAlmostEqual(results[0]['SITI_noref_feature_ti_score'], 13.580963712032636)

        self.assertAlmostEqual(results[1]['SITI_noref_feature_si_score'], 83.46568284569439)
        self.assertAlmostEqual(results[1]['SITI_noref_feature_ti_score'], 15.570422677885475)

        perc75 = partial(np.percentile, q=75)
        [result.set_score_aggregate_method(perc75) for result in results]

        self.assertAlmostEqual(results[0]['SITI_noref_feature_si_score'], 80.01507800178135)
        self.assertAlmostEqual(results[0]['SITI_noref_feature_ti_score'], 14.71399321502522)

        self.assertAlmostEqual(results[1]['SITI_noref_feature_si_score'], 84.00886573293045)
        self.assertAlmostEqual(results[1]['SITI_noref_feature_ti_score'], 16.862281850395433)

        perc100 = partial(np.percentile, q=100)
        [result.set_score_aggregate_method(perc100) for result in results]

        self.assertAlmostEqual(results[0]['SITI_noref_feature_si_score'], 82.84005099011819)
        self.assertAlmostEqual(results[0]['SITI_noref_feature_ti_score'], 17.60733889303054)

        self.assertAlmostEqual(results[1]['SITI_noref_feature_si_score'], 85.28055474796807)
        self.assertAlmostEqual(results[1]['SITI_noref_feature_ti_score'], 19.500837587311423)

    def test_noref_moment_fextractor_proc(self):

        ref_path, dis_path, asset, asset_original = set_default_576_324_videos_for_testing()

        callback_dict = {
            'dis_proc_callback': 'identity',
        }
        asset.asset_dict.update(callback_dict)
        asset_original.asset_dict.update(callback_dict)

        self.fextractor = MomentNorefFeatureExtractor(
            [asset, asset_original],
            None, fifo_mode=True,
            result_store=None
        )
        self.fextractor.run(parallelize=True)

        results = self.fextractor.results

        self.assertAlmostEqual(results[0]['Moment_noref_feature_1st_score'], 61.332006624999984)
        self.assertAlmostEqual(results[0]['Moment_noref_feature_2nd_score'], 4798.659574041666)
        self.assertAlmostEqual(results[0]['Moment_noref_feature_var_score'], 1036.8371843488285)

        self.assertAlmostEqual(results[1]['Moment_noref_feature_1st_score'], 59.788567297525134)
        self.assertAlmostEqual(results[1]['Moment_noref_feature_2nd_score'], 4696.668388042271)
        self.assertAlmostEqual(results[1]['Moment_noref_feature_var_score'], 1121.519917231207)


class FeatureExtractorSaveWorkfilesTest(MyTestCase):

    def setUp(self):
        super().setUp()
        self.result_store = FileSystemResultStore()

    def tearDown(self):
        if hasattr(self, 'fextractor'):
            self.fextractor.remove_results()
        super().tearDown()

    def test_noref_moment_fextractor(self):

        ref_path, dis_path, asset, asset_original = set_default_576_324_videos_for_testing()

        self.fextractor = MomentNorefFeatureExtractor(
            [asset, asset_original],
            None, fifo_mode=False,
            result_store=self.result_store,
            save_workfiles=True,
        )
        self.fextractor.run()

        results = self.fextractor.results

        self.assertAlmostEqual(results[0]['Moment_noref_feature_1st_score'], 61.332006624999984)
        self.assertAlmostEqual(results[0]['Moment_noref_feature_2nd_score'], 4798.659574041666)
        self.assertAlmostEqual(results[0]['Moment_noref_feature_var_score'], 1036.8371843488285)

        self.assertAlmostEqual(results[1]['Moment_noref_feature_1st_score'], 59.788567297525134)
        self.assertAlmostEqual(results[1]['Moment_noref_feature_2nd_score'], 4696.668388042271)
        self.assertAlmostEqual(results[1]['Moment_noref_feature_var_score'], 1121.519917231207)

        self.fextractor.run()

    def test_noref_moment_fextractor_save_workfiles_second_time(self):

        ref_path, dis_path, asset, asset_original = set_default_576_324_videos_for_testing()

        self.fextractor = MomentNorefFeatureExtractor(
            [asset, asset_original],
            None, fifo_mode=False,
            result_store=self.result_store,
            save_workfiles=False,
        )
        self.fextractor.run()

        results = self.fextractor.results

        self.assertAlmostEqual(results[0]['Moment_noref_feature_1st_score'], 61.332006624999984)
        self.assertAlmostEqual(results[0]['Moment_noref_feature_2nd_score'], 4798.659574041666)
        self.assertAlmostEqual(results[0]['Moment_noref_feature_var_score'], 1036.8371843488285)

        self.assertAlmostEqual(results[1]['Moment_noref_feature_1st_score'], 59.788567297525134)
        self.assertAlmostEqual(results[1]['Moment_noref_feature_2nd_score'], 4696.668388042271)
        self.assertAlmostEqual(results[1]['Moment_noref_feature_var_score'], 1121.519917231207)

        self.fextractor = MomentNorefFeatureExtractor(
            [asset, asset_original],
            None, fifo_mode=False,
            result_store=self.result_store,
            save_workfiles=True,
        )
        self.fextractor.run()


if __name__ == '__main__':
    unittest.main(verbosity=2)



#!/usr/bin/env python3
import os
import sys
from vmaf.core.quality_runner import QualityRunner
from vmaf.core.result_store import FileSystemResultStore
from vmaf.routine import run_remove_results_for_dataset
from vmaf.tools.misc import import_python_file

__copyright__ = "Copyright 2016-2020, Netflix, Inc."
__license__ = "BSD+Patent"


def print_usage():
    quality_runner_types = ['VMAF', 'PSNR', 'SSIM', 'MS_SSIM']
    print("usage: " + os.path.basename(sys.argv[0]) + \
          " quality_type dataset_filepath\n")
    print("quality_type:\n\t" + "\n\t".join(quality_runner_types) +"\n")


def main():
    if len(sys.argv) < 3:
        print_usage()
        return 2

    try:
        quality_type = sys.argv[1]
        dataset_filepath = sys.argv[2]
    except ValueError:
        print_usage()
        return 2

    try:
        dataset = import_python_file(dataset_filepath)
    except Exception as e:
        print("Error: " + str(e))
        return 1

    try:
        runner_class = QualityRunner.find_subclass(quality_type)
    except:
        print_usage()
        return 2

    result_store = FileSystemResultStore()

    run_remove_results_for_dataset(result_store, dataset, runner_class)

    return 0


if __name__ == '__main__':
    ret = main()
    exit(ret)
